"""Extract Cherenkov photons information from a CORSIKA IACT file."""

import functools
import logging
import operator
import re
import time
from pathlib import Path

import boost_histogram as bh
import numpy as np
from astropy import units as u
from astropy.io.misc import yaml
from astropy.units import cds
from corsikaio.subblocks import event_header, get_units_from_fields, run_header
from ctapipe.io import write_table
from eventio import IACTFile

from simtools import version
from simtools.io import io_handler
from simtools.io.ascii_handler import collect_data_from_file
from simtools.io.hdf5_handler import fill_hdf5_table
from simtools.utils.geometry import convert_2d_to_radial_distr, rotate
from simtools.utils.names import sanitize_name
from simtools.visualization import plot_corsika_histograms as visualize

X_AXIS_STRING = "x axis"
Y_AXIS_STRING = "y axis"
Z_AXIS_STRING = "z axis"


class HistogramNotCreatedError(Exception):
    """Exception for histogram not created."""


class CorsikaHistograms:
    """
    Extracts the Cherenkov photons information from a CORSIKA IACT file.

    Parameters
    ----------
    input_file: str or Path
        CORSIKA IACT file provided by the CORSIKA simulation.
    label: str
        Instance label.
    output_path: str
        Path where to save the output of the class methods.
    hdf5_file_name: str
        HDF5 file name for histogram storage.

    Raises
    ------
    FileNotFoundError:
        if the input file given does not exist.
    """

    def __init__(self, input_file, label=None, output_path=None, hdf5_file_name=None):
        self.label = label
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init CorsikaHistograms")
        self.input_file = input_file

        self.input_file = Path(self.input_file)
        if not self.input_file.exists():
            msg = f"file {self.input_file} does not exist."
            self._logger.error(msg)
            raise FileNotFoundError

        self.io_handler = io_handler.IOHandler()
        _default_output_path = self.io_handler.get_output_directory("corsika")

        if output_path is None:
            self.output_path = _default_output_path
        else:
            self.output_path = Path(output_path)

        if hdf5_file_name is None:
            self.hdf5_file_name = re.split(r"\.", self.input_file.name)[0] + ".hdf5"
        else:
            self.hdf5_file_name = hdf5_file_name

        self._telescope_indices = None
        self._telescope_positions = None
        self.num_events = None
        self.num_of_hist = None
        self.num_telescopes = None
        self._num_photons_per_event_per_telescope = None
        self._num_photons_per_event = None
        self._num_photons_per_telescope = None
        self.__meta_dict = None
        self._dict_2d_distributions = None
        self._dict_1d_distributions = None
        self._event_azimuth_angles = None
        self._event_zenith_angles = None
        self._hist_config = None
        self._total_num_photons = None
        self._magnetic_field_x = None
        self._magnetic_field_z = None
        self._event_total_energies = None
        self._event_first_interaction_heights = None
        self._corsika_version = None
        self.event_information = None
        self._individual_telescopes = None
        self._allowed_histograms = {"hist_position", "hist_direction", "hist_time_altitude"}
        self._allowed_1d_labels = {"wavelength", "time", "altitude"}
        self._allowed_2d_labels = {"counts", "density", "direction", "time_altitude"}
        self._header = None
        self.hist_position = None
        self.hist_direction = None
        self.hist_time_altitude = None

        self.read_event_information()
        self._initialize_header()

    def parse_telescope_indices(self, indices_arg):
        """Return telescope indices as ndarray[int] or None.

        Accepts None, a sequence of strings/ints. Raises ValueError on invalid input.
        """
        if indices_arg is None:
            return None
        try:
            return np.array(indices_arg).astype(int)
        except ValueError as exc:
            msg = (
                f"{indices_arg} not a valid input. Please use integer numbers for telescope_indices"
            )
            self._logger.error(msg)
            raise ValueError(msg) from exc

    def should_overwrite(
        self, write_hdf5: bool, event1d: list | None, event2d: list | None
    ) -> bool:
        """Return True if output HDF5 exists and any writing flag is requested."""
        exists = Path(self.hdf5_file_name).exists()
        if exists and (write_hdf5 or bool(event1d) or bool(event2d)):
            self._logger.warning(
                f"Output hdf5 file {self.hdf5_file_name} already exists. Overwriting it."
            )
            return True
        return False

    def run_export_pipeline(
        self,
        *,
        individual_telescopes: bool,
        hist_config,
        indices_arg,
        write_pdf: bool,
        write_hdf5: bool,
        event1d: list | None,
        event2d: list | None,
        test: bool = False,
    ) -> dict:
        """Run the full histogram export pipeline and return output artifact paths.

        Returns a dict with optional keys: pdf_photons, pdf_event1d, pdf_event2d.
        """
        outputs: dict[str, Path | None] = {
            "pdf_photons": None,
            "pdf_event_1d": None,
            "pdf_event_2d": None,
        }

        indices = self.parse_telescope_indices(indices_arg)
        overwrite = self.should_overwrite(write_hdf5, event1d, event2d)

        self.set_histograms(
            telescope_indices=indices,
            individual_telescopes=individual_telescopes,
            hist_config=hist_config,
        )

        if write_pdf:
            pdf_path = visualize.export_all_photon_figures_pdf(self, test=test)
            outputs["pdf_photons"] = pdf_path
        if write_hdf5:
            self.export_histograms(overwrite=overwrite)

        if event1d is not None:
            outputs["pdf_event_1d"] = visualize.derive_event_1d_histograms(
                self, event1d, pdf=write_pdf, hdf5=write_hdf5, overwrite=not write_hdf5
            )
        if event2d is not None:
            outputs["pdf_event_2d"] = visualize.derive_event_2d_histograms(
                self,
                event2d,
                pdf=write_pdf,
                hdf5=write_hdf5,
                overwrite=not (write_hdf5 or bool(event1d)),
            )

        return outputs

    @property
    def hdf5_file_name(self):
        """
        Property for the hdf5 file name.

        The idea of this property is to allow setting (or changing) the name of the hdf5 file
        even after creating the class instance.
        """
        return self._hdf5_file_name

    @hdf5_file_name.setter
    def hdf5_file_name(self, hdf5_file_name):
        """
        Set the hdf5_file_name to the argument passed.

        Parameters
        ----------
        hdf5_file_name: str
            The name of hdf5 file to be set.
        """
        self._hdf5_file_name = Path(self.output_path).joinpath(hdf5_file_name).absolute().as_posix()

    @property
    def corsika_version(self):
        """
        Get the version of the CORSIKA IACT file.

        Returns
        -------
        float:
            The version of CORSIKA used to produce the CORSIKA IACT file given by self.input_file.
        """
        if self._corsika_version is None:
            all_corsika_versions = list(run_header.run_header_types.keys())
            header = list(self.iact_file.header)

            for i_version in reversed(all_corsika_versions):
                # Get the event header for this software version being tested.
                single_run_header = run_header.run_header_types[i_version]
                # Get the position in the dictionary, where the version is.
                version_index_position = np.argwhere(
                    np.array(list(single_run_header.names)) == "version"
                )[0]

                # Check if version tested is the same as the version written in the file header.
                if i_version == np.trunc(float(header[version_index_position[0]]) * 10) / 10:
                    # If the version found is the same as the initial guess, leave the loop,
                    # otherwise, iterate until we find the correct version.
                    self._corsika_version = np.around(float(header[version_index_position[0]]), 3)
                    break
        return self._corsika_version

    def _initialize_header(self):
        """Initialize the header."""
        self.all_run_keys = list(
            run_header.run_header_types[np.around(self.corsika_version, 1)].names
        )
        self._header = {}

        # Get units of the header
        all_run_units = get_units_from_fields(
            run_header.run_header_fields[np.trunc(self.corsika_version * 10) / 10]
        )
        all_header_astropy_units = self._get_header_astropy_units(self.all_run_keys, all_run_units)

        # Fill the header dictionary
        for i_key, key in enumerate(self.all_run_keys[1:]):  # starting at the second
            # element to avoid the non-numeric key.
            self._header[key] = self.iact_file.header[i_key + 1] * all_header_astropy_units[key]

    @property
    def header(self):
        """
        Get the run header.

        Returns
        -------
        dict:
            The run header.
        """
        return self._header

    def read_event_information(self):
        """
        Read the information about the events from their headers and save as a class instance.

        The main information can also be fetched individually through the functions below.
        For the remaining information (such as px, py, pz), use this function.

        """
        if self.event_information is None:
            with IACTFile(self.input_file) as self.iact_file:
                self.telescope_positions = np.array(self.iact_file.telescope_positions)
                self.num_telescopes = np.size(self.telescope_positions, axis=0)
                self.all_event_keys = list(
                    event_header.event_header_types[np.trunc(self.corsika_version * 10) / 10].names
                )
                all_event_units = get_units_from_fields(
                    event_header.event_header_fields[np.trunc(self.corsika_version * 10) / 10]
                )

                self.event_information = {key: [] for key in self.all_event_keys}

                self.num_events = 0
                # Build a dictionary with the parameters for the events.
                for event in self.iact_file:
                    for i_key, key in enumerate(self.all_event_keys[1:]):
                        self.event_information[key].append(event.header[i_key + 1])

                    self.num_events += 1

                all_event_astropy_units = self._get_header_astropy_units(
                    self.all_event_keys, all_event_units
                )

                # Add the units to dictionary with the parameters and turn it into
                # astropy.Quantities.
                for i_key, key in enumerate(self.all_event_keys[1:]):  # starting at the second
                    # element to avoid the non-numeric (e.g. 'EVTH') key.
                    self.event_information[key] = (
                        np.array(self.event_information[key]) * all_event_astropy_units[key]
                    )

    def _get_header_astropy_units(self, parameters, non_astropy_units):
        """
        Return the dictionary with astropy units from the given list of parameters.

        Parameters
        ----------
        parameters: list
            The list of parameters to extract the astropy units.
        non_astropy_units: dict
            A dictionary with the parameter units (in strings).

        Returns
        -------
        dict:
            A dictionary with the astropy units.
        """
        # Build a dictionary with astropy units for the unit of the event's (header's) parameters.
        all_event_astropy_units = {}
        for key in parameters[1:]:  # starting at the second
            # element to avoid the non-numeric (e.g. 'EVTH') key.

            # We extract the astropy unit (dimensionless in case no unit is provided).
            if key in non_astropy_units:
                with cds.enable():
                    unit = u.Unit(non_astropy_units[key])
            else:
                unit = u.dimensionless_unscaled
            all_event_astropy_units[key] = unit
        return all_event_astropy_units

    @property
    def telescope_indices(self):
        """
        The telescope index (or indices), which are considered for the production of the histograms.

        Returns
        -------
        list:
            The indices of the telescopes of interest.
        """
        return self._telescope_indices

    @telescope_indices.setter
    def telescope_indices(self, telescope_new_indices):
        """
        Set the telescope index (or indices).

        If self.individual_telescopes is True, the indices of the telescopes passed are analyzed
        individually (different histograms for each telescope) even if all telescopes are listed.

        Parameters
        ----------
        telescope_new_indices: int or list of int or np.array of int
            The indices of the specific telescopes to be inspected. If not specified, all telescopes
            are treated together in one histogram and the value of self._telescope_indices is a list
            of all telescope indices.

        Raises
        ------
        TypeError:
            if the indices passed through telescope_index are not of type int.
        """
        if telescope_new_indices is None:
            self._telescope_indices = np.arange(self.num_telescopes)
        else:
            if not isinstance(telescope_new_indices, list | np.ndarray):
                telescope_new_indices = np.array([telescope_new_indices])
            for i_telescope in telescope_new_indices:
                if not isinstance(i_telescope, int | np.integer):
                    msg = "The index or indices given are not of type int."
                    self._logger.error(msg)
                    raise TypeError
            self._telescope_indices = np.sort(telescope_new_indices)

    @property
    def hist_config(self):
        """
        The configuration of the histograms.

        Returns
        -------
        dict:
            the dictionary with the histogram configuration.
        """
        if self._hist_config is None:
            msg = (
                "No histogram configuration was defined before. The default config is being "
                "created now."
            )
            self._logger.warning(msg)
            self._hist_config = self._create_histogram_default_config()
        return self._hist_config

    @hist_config.setter
    def hist_config(self, input_config):
        """
        Set the configuration for the histograms (e.g., bin size, min and max values, etc).

        The input is allowed either through a yaml file or a dictionary. If nothing is given,
        the dictionary is created with default values.

        Parameters
        ----------
        input_config: str, Path, dict or NoneType
            yaml file with the configuration parameters to create the histograms. For the correct
            format, please look at the docstring at _create_histogram_default_config.
            Alternatively, it can be a dictionary with the configuration parameters to create
            the histograms.
        """
        if isinstance(input_config, dict):
            self._hist_config = input_config
        else:
            self._hist_config = collect_data_from_file(input_config) if input_config else None

    def hist_config_to_yaml(self, file_name=None):
        """
        Save the histogram configuration dictionary to a yaml file.

        Parameters
        ----------
        file_name: str
            Name of the output file, in which to save the histogram configuration.

        """
        if file_name is None:
            file_name = "hist_config"
        file_name = Path(file_name).with_suffix(".yml")
        output_config_file_name = Path(self.output_path).joinpath(file_name)
        with open(output_config_file_name, "w", encoding="utf-8") as file:
            yaml.dump(self.hist_config, file)

    def _create_histogram_default_config(self):
        """
        Create a dictionary with the configuration necessary to create the histograms.

        It is used only in case the configuration is not provided in a yaml file or dict.

        Three histograms are created: hist_position with 3 dimensions (x, y positions and the
        wavelength), hist_direction with 2 dimensions (direction cosines in x and y directions),
        hist_time_altitude with 2 dimensions (time and altitude of emission).

        Four arguments are passed to each dimension in the dictionary:

        "bins": the number of bins,
        "start": the first element of the histogram,
        "stop": the last element of the histogram, and
        "scale" to define the scale of the bins which can be "linear" or "log". If "log", the
         common logarithm (log10) is applied to the axes. The start and stop values have to be
         valid, i.e., >0.

        Returns
        -------
        dict:
            Dictionary with the configuration parameters to create the histograms.
        """
        if self.individual_telescopes is False:
            xy_maximum = 1000 * u.m
            xy_bin = 100

        else:
            xy_maximum = 16 * u.m
            xy_bin = 64
        return {
            "hist_position": {
                X_AXIS_STRING: {
                    "bins": xy_bin,
                    "start": -xy_maximum,
                    "stop": xy_maximum,
                    "scale": "linear",
                },
                Y_AXIS_STRING: {
                    "bins": xy_bin,
                    "start": -xy_maximum,
                    "stop": xy_maximum,
                    "scale": "linear",
                },
                Z_AXIS_STRING: {
                    "bins": 80,
                    "start": 200 * u.nm,
                    "stop": 1000 * u.nm,
                    "scale": "linear",
                },
            },
            "hist_direction": {
                X_AXIS_STRING: {
                    "bins": 100,
                    "start": -1,
                    "stop": 1,
                    "scale": "linear",
                },
                Y_AXIS_STRING: {
                    "bins": 100,
                    "start": -1,
                    "stop": 1,
                    "scale": "linear",
                },
            },
            "hist_time_altitude": {
                X_AXIS_STRING: {
                    "bins": 100,
                    "start": -2000 * u.ns,
                    "stop": 2000 * u.ns,
                    "scale": "linear",
                },
                Y_AXIS_STRING: {
                    "bins": 100,
                    "start": 120 * u.km,
                    "stop": 0 * u.km,
                    "scale": "linear",
                },
            },
        }

    def _create_regular_axes(self, label):
        """
        Create regular axis for histograms.

        Parameters
        ----------
        label: str
            Label to identify to which histogram the new axis belongs.

        Raises
        ------
        ValueError:
            if label is not valid.
        """
        transform = {"log": bh.axis.transform.log, "linear": None}

        if label not in self._allowed_histograms:
            msg = f"allowed labels must be one of the following: {self._allowed_histograms}"
            self._logger.error(msg)
            raise ValueError(msg)

        all_axes = [X_AXIS_STRING, Y_AXIS_STRING]
        if label == "hist_position":
            all_axes.append(Z_AXIS_STRING)

        boost_axes = []
        for axis in all_axes:
            if isinstance(self.hist_config[label][axis]["start"], u.quantity.Quantity):
                start = self.hist_config[label][axis]["start"].value
                stop = self.hist_config[label][axis]["stop"].value
            else:
                start = self.hist_config[label][axis]["start"]
                stop = self.hist_config[label][axis]["stop"]
            boost_axes.append(
                bh.axis.Regular(
                    bins=self.hist_config[label][axis]["bins"],
                    start=start,
                    stop=stop,
                    transform=transform[self.hist_config[label][axis]["scale"]],
                )
            )
        return boost_axes

    def _create_histograms(self, individual_telescopes=False):
        """
        Create the histogram instances.

        Parameters
        ----------
        individual_telescopes: bool
            if False, the histograms are filled for all given telescopes together.
            if True, one histogram is set for each telescope separately.
        """
        self.individual_telescopes = individual_telescopes
        self.num_of_hist = len(self.telescope_indices) if self.individual_telescopes is True else 1

        self.hist_position, self.hist_direction, self.hist_time_altitude = [], [], []

        for _ in range(self.num_of_hist):
            boost_axes_position = self._create_regular_axes("hist_position")
            self.hist_position.append(
                bh.Histogram(
                    boost_axes_position[0],
                    boost_axes_position[1],
                    boost_axes_position[2],
                )
            )
            boost_axes_direction = self._create_regular_axes("hist_direction")
            self.hist_direction.append(
                bh.Histogram(
                    boost_axes_direction[0],
                    boost_axes_direction[1],
                )
            )

            boost_axes_time_altitude = self._create_regular_axes("hist_time_altitude")
            self.hist_time_altitude.append(
                bh.Histogram(
                    boost_axes_time_altitude[0],
                    boost_axes_time_altitude[1],
                )
            )

    def _fill_histograms(self, photons, rotation_around_z_axis=None, rotation_around_y_axis=None):
        """
        Fill histograms with the information of the photons on the ground.

        if the azimuth and zenith angles are provided, the Cherenkov photon's coordinates are
        filled in the plane perpendicular to the incoming direction of the particle.

        Parameters
        ----------
        photons: list
            List of size M of numpy.array of size (N,8), where M is the number of telescopes in the
            array, N is the number of photons that reached each telescope. The following information
             of the Cherenkov photons on the ground are saved:
             x: x position on the ground (CORSIKA coordinate system),
             y: y position on the ground (CORSIKA coordinate system),
             cx: direction cosine in the x direction, i.e., the cosine of the angle between the
             incoming direction and the x axis,
             cy: direction cosine in the y direction, i.e., the cosine of the angle between the
             incoming direction and the y axis,
             time: time of arrival of the photon in ns. The clock starts when the particle crosses
             the top of the atmosphere (CORSIKA-defined) if self.event_first_interaction_heights
             is positive or at first interaction if otherwise.
             zem: altitude where the photon was generated in cm,
             photons: number of photons associated to this bunch,
             wavelength: the wavelength of the photons in nm.
        rotation_around_z_axis: astropy.Quantity
            Angle to rotate the observational plane around the Z axis.
            It can be passed in radians or degrees.
            If not given, no rotation is performed.
        rotation_around_y_axis: astropy.Quantity
            Angle to rotate the observational plane around the Y axis.
            It can be passed in radians or degrees.
            If not given, no rotation is performed.
            rotation_around_z_axis and rotation_around_y_axis are used to align the observational
            plane normal to the incoming direction of the shower particles for the individual
             telescopes (useful for fixed targets).

        Raises
        ------
        IndexError:
            If the index or indices passed though telescope_index are out of range.
        """
        hist_num = 0
        for i_tel_info, photons_info in np.array(
            list(zip(self.telescope_positions, photons)), dtype=object
        )[self.telescope_indices]:
            if rotation_around_z_axis is None or rotation_around_y_axis is None:
                photon_x, photon_y = photons_info["x"], photons_info["y"]
            else:
                photon_x, photon_y = rotate(
                    photons_info["x"],
                    photons_info["y"],
                    rotation_around_z_axis,
                    rotation_around_y_axis,
                )

            if self.individual_telescopes is False:
                # Adding the position of the telescopes to the relative position of the photons
                # such that we have a common coordinate system.
                photon_x = -i_tel_info["x"] + photon_x
                photon_y = -i_tel_info["y"] + photon_y

            self.hist_position[hist_num].fill(
                (photon_x * u.cm).to(u.m),
                (photon_y * u.cm).to(u.m),
                np.abs(photons_info["wavelength"]) * u.nm,
            )

            self.hist_direction[hist_num].fill(photons_info["cx"], photons_info["cy"])
            self.hist_time_altitude[hist_num].fill(
                photons_info["time"] * u.ns, (photons_info["zem"] * u.cm).to(u.km)
            )
            if self.individual_telescopes is True:
                hist_num += 1

    def set_histograms(self, telescope_indices=None, individual_telescopes=None, hist_config=None):
        """
        Create and fill Cherenkov photons histograms using information from the CORSIKA IACT file.

        Parameters
        ----------
        telescope_indices: int or list of int
            The indices of the specific telescopes to be inspected.
        individual_telescopes: bool
            if False, the histograms are supposed to be filled for all telescopes. Default is False.
            if True, one histogram is set for each telescope separately.
        hist_config:
            yaml file with the configuration parameters to create the histograms. For the correct
            format, please look at the docstring of _create_histogram_default_config.
            Alternatively, it can be a dictionary with the configuration parameters to create
            the histograms.

        Returns
        -------
        list: list of boost_histogram.Histogram instances.

        Raises
        ------
        AttributeError:
            if event has not photon saved.
        """
        self.telescope_indices = telescope_indices
        self.individual_telescopes = individual_telescopes
        self.hist_config = hist_config
        self._create_histograms(individual_telescopes=self.individual_telescopes)

        num_photons_per_event_per_telescope_to_set = []
        start_time = time.time()
        self._logger.debug(f"Starting reading the file at {start_time}.")
        with IACTFile(self.input_file) as f:
            event_counter = 0
            for event in f:
                for i_telescope in self.telescope_indices:
                    if hasattr(event, "photon_bunches"):
                        photons = list(event.photon_bunches.values())
                    else:
                        msg = "The event has no associated photon bunches saved. "
                        self._logger.error(msg)
                        raise AttributeError

                    # Count photons only from the telescopes given by self.telescope_indices.
                    num_photons_per_event_per_telescope_to_set.append(event.n_photons[i_telescope])
                self._fill_histograms(
                    photons,
                    self.event_azimuth_angles[event_counter],
                    self.event_zenith_angles[event_counter],
                )
                event_counter += 1
        self.num_photons_per_event_per_telescope = num_photons_per_event_per_telescope_to_set
        self._logger.debug(
            f"Finished reading the file and creating the histograms in {time.time() - start_time} "
            f"seconds"
        )

    @property
    def individual_telescopes(self):
        """Return the individual telescopes as property."""
        return self._individual_telescopes

    @individual_telescopes.setter
    def individual_telescopes(self, new_individual_telescopes: bool):
        """
        Define individual telescopes.

        Parameters
        ----------
        new_individual_telescopes: bool
            if False, the histograms are supposed to be filled for all telescopes.
            if True, one histogram is set for each telescope separately.
        """
        if new_individual_telescopes is None:
            self._individual_telescopes = False
        else:
            self._individual_telescopes = new_individual_telescopes

    def _raise_if_no_histogram(self):
        """
        Raise an error if the histograms were not created.

        Raises
        ------
        HistogramNotCreatedError:
            if the histogram was not previously created.
        """
        for histogram in self._allowed_histograms:
            if not hasattr(self, histogram) or getattr(self, histogram) is None:
                msg = (
                    "The histograms were not created. Please, use create_histograms to create "
                    "histograms from the CORSIKA output file."
                )
                self._logger.error(msg)
                raise HistogramNotCreatedError

    def _get_hist_2d_projection(self, label):
        """
        Get 2D distributions.

        Parameters
        ----------
        label: str
            Label to indicate which histogram.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            The x bin edges of the histograms.
        numpy.array
            The y bin edges of the histograms.

        Raises
        ------
        ValueError:
            if label is not valid.
        """
        if label not in self._allowed_2d_labels:
            msg = f"label is not valid. Valid entries are {self._allowed_2d_labels}"
            self._logger.error(msg)
            raise ValueError(msg)
        self._raise_if_no_histogram()

        num_telescopes_to_fill = (
            len(self.telescope_indices) if self.individual_telescopes is True else 1
        )

        x_bin_edges, y_bin_edges, hist_values = [], [], []
        for i_telescope in range(num_telescopes_to_fill):
            mini_hist = None
            if label == "counts":
                mini_hist = self.hist_position[i_telescope][:, :, sum]
                hist_values.append(mini_hist.view().T)
            elif label == "density":
                mini_hist = self.hist_position[i_telescope][:, :, sum]
                areas = functools.reduce(operator.mul, mini_hist.axes.widths)
                hist_values.append(mini_hist.view().T / areas)
            elif label == "direction":
                mini_hist = self.hist_direction[i_telescope]
                hist_values.append(self.hist_direction[i_telescope].view().T)
            elif label == "time_altitude":
                mini_hist = self.hist_time_altitude[i_telescope]
                hist_values.append(self.hist_time_altitude[i_telescope].view().T)
            if mini_hist is not None:
                x_bin_edges.append(mini_hist.axes.edges[0].flatten())
                y_bin_edges.append(mini_hist.axes.edges[1].flatten())

        return np.array(hist_values), np.array(x_bin_edges), np.array(y_bin_edges)

    def get_2d_photon_position_distr(self):
        """
        Get 2D histograms of position of the Cherenkov photons on the ground.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            The x bin edges of the count histograms in x, usually in meters.
        numpy.array
            The y bin edges of the count histograms in y, usually in meters.
        """
        return self._get_hist_2d_projection("counts")

    def get_2d_photon_density_distr(self):
        """
        Get 2D histograms of position of the Cherenkov photons on the ground.

        It returns the photon density per square meter.

        Returns
        -------
        numpy.ndarray
            The values of the histogram, usually in $m^{-2}$
        numpy.array
            The x bin edges of the density/count histograms in x, usually in meters.
        numpy.array
            The y bin edges of the density/count histograms in y, usually in meters.
        """
        return self._get_hist_2d_projection("density")

    def get_2d_photon_direction_distr(self):
        """
        Get 2D histograms of incoming direction of the Cherenkov photons on the ground.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            The x bin edges of the direction histograms in cos(x).
        numpy.array
            The y bin edges of the direction histograms in cos(y)
        """
        return self._get_hist_2d_projection("direction")

    def get_2d_photon_time_altitude_distr(self):
        """
        Get 2D histograms of the time and altitude of the photon production.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            The x bin edges of the time_altitude histograms, usually in ns.
        numpy.array
            The y bin edges of the time_altitude histograms, usually in km.
        """
        return self._get_hist_2d_projection("time_altitude")

    def get_2d_num_photons_distr(self):
        """
        Get the distribution of Cherenkov photons per event per telescope.

        It returns the 2D array accounting for the events from the telescopes given
        by self.telescope_indices.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            An array that counts the telescopes in self.telescope_indices
        numpy.array
            Number of photons per event per telescope in self.telescope_indices.
        """
        num_events_array = np.arange(self.num_events + 1).reshape(1, self.num_events + 1)
        # It counts only the telescope indices given by self.telescope_indices.
        # The + 1 closes the last edge.
        telescope_counter = np.arange(len(self.telescope_indices) + 1).reshape(
            1, len(self.telescope_indices) + 1
        )
        hist_2d = np.array(self.num_photons_per_event_per_telescope)
        hist_2d = hist_2d.reshape((1, len(self.telescope_indices), self.num_events))
        return (hist_2d, num_events_array, telescope_counter)

    def _get_hist_1d_projection(self, label):
        """
        Get 1D distributions.

        Parameters
        ----------
        label: str
            Label to indicate which histogram.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            The bin edges of the histogram.

        Raises
        ------
        ValueError:
            if label is not valid.
        """
        if label not in self._allowed_1d_labels:
            msg = f"{label} is not valid. Valid entries are {self._allowed_1d_labels}"
            self._logger.error(msg)
            raise ValueError(msg)
        self._raise_if_no_histogram()

        x_bin_edges_list, hist_1d_list = [], []
        for i_hist, _ in enumerate(self.hist_position):
            mini_hist = None
            if label == "wavelength":
                mini_hist = self.hist_position[i_hist][sum, sum, :]
            elif label == "time":
                mini_hist = self.hist_time_altitude[i_hist][:, sum]
            elif label == "altitude":
                mini_hist = self.hist_time_altitude[i_hist][sum, :]

            x_bin_edges_list.append(mini_hist.axes.edges.T.flatten()[0])
            hist_1d_list.append(mini_hist.view().T)
        return np.array(hist_1d_list), np.array(x_bin_edges_list)

    def _get_bins_max_dist(self, bins=None, max_dist=None):
        """
        Get the number of bins and the max distance to generate the radial and density histograms.

        Parameters
        ----------
        bins: float
            Number of bins of the radial distribution.
        max_dist: float
            Maximum distance to consider in the 1D histogram (in meters).
        """
        hist_position = "hist_position"
        if max_dist is None:
            max_dist = np.amax(
                [
                    self.hist_config[hist_position][X_AXIS_STRING]["start"].to(u.m).value,
                    self.hist_config[hist_position][X_AXIS_STRING]["stop"].to(u.m).value,
                    self.hist_config[hist_position][Y_AXIS_STRING]["start"].to(u.m).value,
                    self.hist_config[hist_position][Y_AXIS_STRING]["stop"].to(u.m).value,
                ]
            )
        if bins is None:
            bins = (
                np.amax(
                    [
                        self.hist_config[hist_position][X_AXIS_STRING]["bins"],
                        self.hist_config[hist_position][Y_AXIS_STRING]["bins"],
                    ]
                )
                // 2
            )  # //2 because of the 2D array going into the negative and
            # positive axis
        return bins, max_dist

    def get_photon_radial_distr(self, bins=None, max_dist=None):
        """
        Get the phton radial distribution on the ground in relation to the center of the array.

        Parameters
        ----------
        bins: float
            Number of bins of the radial distribution.
        max_dist: float
            Maximum distance to consider in the 1D histogram (in meters).

        Returns
        -------
        np.array
            The counts of the 1D histogram with size = int(max_dist/bin_size).
        np.array
            The bin edges of the 1D histogram in meters with size = int(max_dist/bin_size) + 1,
            usually in meter.
        """
        bins, max_dist = self._get_bins_max_dist(bins=bins, max_dist=max_dist)
        bin_edges_1d_list, hist_1d_list = [], []

        hist_2d_values_list, x_position_list, y_position_list = self.get_2d_photon_position_distr()

        for i_hist, x_pos in enumerate(x_position_list):
            hist_1d, bin_edges_1d = convert_2d_to_radial_distr(
                hist_2d_values_list[i_hist],
                x_pos,
                y_position_list[i_hist],
                bins=bins,
                max_dist=max_dist,
            )
            bin_edges_1d_list.append(bin_edges_1d)
            hist_1d_list.append(hist_1d)
        return np.array(hist_1d_list), np.array(bin_edges_1d_list)

    def get_photon_density_distr(self, bins=None, max_dist=None):
        """
        Get the photon density distribution on the ground in relation to the center of the array.

        Parameters
        ----------
        bins: float
            Number of bins of the radial distribution.
        max_dist: float
            Maximum distance to consider in the 1D histogram (in meters).

        Returns
        -------
        np.array
            The density distribution of the 1D histogram with size = int(max_dist/bin_size),
            usually in $m^{-2}$.
        np.array
            The bin edges of the 1D histogram in meters with size = int(max_dist/bin_size) + 1,
            usually in meter.
        """
        bins, max_dist = self._get_bins_max_dist(bins=bins, max_dist=max_dist)
        bin_edges_1d_list, hist_1d_list = [], []

        hist_2d_values_list, x_position_list, y_position_list = self.get_2d_photon_density_distr()

        for i_hist, _ in enumerate(x_position_list):
            hist_1d, bin_edges_1d = convert_2d_to_radial_distr(
                hist_2d_values_list[i_hist],
                x_position_list[i_hist],  # pylint: disable=unnecessary-list-index-lookup
                y_position_list[i_hist],
                bins=bins,
                max_dist=max_dist,
            )
            bin_edges_1d_list.append(bin_edges_1d)
            hist_1d_list.append(hist_1d)
        return np.array(hist_1d_list), np.array(bin_edges_1d_list)

    def get_photon_wavelength_distr(self):
        """
        Get histograms with the wavelengths of the photon bunches.

        Returns
        -------
        np.array
            The counts of the wavelength histogram.
        np.array
            The bin edges of the wavelength histogram in nanometers.

        """
        return self._get_hist_1d_projection("wavelength")

    def get_photon_time_of_emission_distr(self):
        """
        Get the distribution of the emitted time of the Cherenkov photons.

        The clock starts when the particle crosses the top of the atmosphere (CORSIKA-defined) if
        self.event_first_interaction_heights is positive or at first interaction if otherwise.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            The bin edges of the time histograms in ns.

        """
        return self._get_hist_1d_projection("time")

    def get_photon_altitude_distr(self):
        """
        Get the emission altitude of the Cherenkov photons.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            The bin edges of the photon altitude histograms in km.

        """
        return self._get_hist_1d_projection("altitude")

    @property
    def num_photons_per_event_per_telescope(self):
        """The number of photons per event per telescope."""
        return self._num_photons_per_event_per_telescope

    @num_photons_per_event_per_telescope.setter
    def num_photons_per_event_per_telescope(self, num_photons_per_event_per_telescope_to_set):
        """Set the number of photons per event per telescope."""
        self._num_photons_per_event_per_telescope = (
            np.array(num_photons_per_event_per_telescope_to_set)
            .reshape(self.num_events, len(self.telescope_indices))
            .T
        )

    @property
    def num_photons_per_event(self):
        """
        Get the the number of photons per events.

        Includes the telescopes indicated by self.telescope_indices.

        Returns
        -------
        numpy.array
            Number of photons per event.
        """
        self._num_photons_per_event = np.sum(self.num_photons_per_event_per_telescope, axis=0)
        return self._num_photons_per_event

    def get_num_photons_per_event_distr(self, bins=50, hist_range=None):
        """
        Get the distribution of photons per event.

        Parameters
        ----------
        bins: float
            Number of bins for the histogram.
        hist_range: 2-tuple
            Tuple to define the range of the histogram.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            Number of photons per event.
        """
        hist, bin_edges = np.histogram(self.num_photons_per_event, bins=bins, range=hist_range)
        return hist.reshape(1, bins), bin_edges.reshape(1, bins + 1)

    def get_num_photons_per_telescope_distr(self, bins=50, hist_range=None):
        """
        Get the distribution of photons per telescope.

        Parameters
        ----------
        bins: float
            Number of bins for the histogram.
        hist_range: 2-tuple
            Tuple to define the range of the histogram.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            Number of photons per telescope.
        """
        hist, bin_edges = np.histogram(self.num_photons_per_telescope, bins=bins, range=hist_range)
        return hist.reshape(1, bins), bin_edges.reshape(1, bins + 1)

    def export_histograms(self, overwrite=False):
        """
        Export the histograms to hdf5 files.

        Parameters
        ----------
        overwrite: bool
            If True overwrites the histograms already saved in the hdf5 file.
        """
        self._export_1d_histograms(overwrite=overwrite)
        self._export_2d_histograms(overwrite=False)

    @property
    def _meta_dict(self):
        """
        Define the meta dictionary for exporting the histograms.

        Returns
        -------
        dict
            Meta dictionary for the hdf5 files with the histograms.
        """
        if self.__meta_dict is None:
            self.__meta_dict = {
                "corsika_version": self.corsika_version,
                "simtools_version": version.__version__,
                "iact_file": self.input_file.name,
                "telescope_indices": list(self.telescope_indices),
                "individual_telescopes": self.individual_telescopes,
                "note": "Only lower bin edges are given.",
            }
        return self.__meta_dict

    @property
    def dict_1d_distributions(self):
        """
        Dictionary to label the 1D distributions according to the class methods.

        Returns
        -------
        dict:
            The dictionary with information about the 1D distributions.
        """
        fn_key = "function"
        file_name = "file name"
        title = "title"
        bin_edges = "bin edges"
        axis_unit = "axis unit"
        self._dict_1d_distributions = {
            "wavelength": {
                fn_key: "get_photon_wavelength_distr",
                file_name: "hist_1d_photon_wavelength_distr",
                title: "Photon wavelength distribution",
                bin_edges: "wavelength",
                axis_unit: self.hist_config["hist_position"][Z_AXIS_STRING]["start"].unit,
            },
            "counts": {
                fn_key: "get_photon_radial_distr",
                file_name: "hist_1d_photon_radial_distr",
                title: "Radial photon distribution on the ground",
                bin_edges: "Distance to center",
                axis_unit: self.hist_config["hist_position"][X_AXIS_STRING]["start"].unit,
            },
            "density": {
                fn_key: "get_photon_density_distr",
                file_name: "hist_1d_photon_density_distr",
                title: "Photon density distribution on the ground",
                bin_edges: "Distance to center",
                axis_unit: self.hist_config["hist_position"][X_AXIS_STRING]["start"].unit,
            },
            "time": {
                fn_key: "get_photon_time_of_emission_distr",
                file_name: "hist_1d_photon_time_distr",
                title: "Photon time of arrival distribution",
                bin_edges: "Time of arrival",
                axis_unit: self.hist_config["hist_time_altitude"][X_AXIS_STRING]["start"].unit,
            },
            "altitude": {
                fn_key: "get_photon_altitude_distr",
                file_name: "hist_1d_photon_altitude_distr",
                title: "Photon altitude of emission distribution",
                bin_edges: "Altitude of emission",
                axis_unit: self.hist_config["hist_time_altitude"][Y_AXIS_STRING]["start"].unit,
            },
            "num_photons_per_event": {
                fn_key: "get_num_photons_per_event_distr",
                file_name: "hist_1d_photon_per_event_distr",
                title: "Photons per event distribution",
                bin_edges: "Event counter",
                axis_unit: u.dimensionless_unscaled,
            },
            "num_photons_per_telescope": {
                fn_key: "get_num_photons_per_telescope_distr",
                file_name: "hist_1d_photon_per_telescope_distr",
                title: "Photons per telescope distribution",
                bin_edges: "Telescope counter",
                axis_unit: u.dimensionless_unscaled,
            },
        }
        return self._dict_1d_distributions

    def _export_1d_histograms(self, overwrite=False):
        """
        Auxiliary function to export only the 1D histograms.

        Parameters
        ----------
        overwrite: bool
            If True overwrites the histograms already saved in the hdf5 file.
        """
        axis_unit = "axis unit"
        for _, function_dict in self.dict_1d_distributions.items():
            self._meta_dict["Title"] = sanitize_name(function_dict["title"])
            histogram_function = getattr(self, function_dict["function"])
            hist_1d_list, x_bin_edges_list = histogram_function()
            x_bin_edges_list = x_bin_edges_list * function_dict[axis_unit]
            if function_dict["function"] == "get_photon_density_distr":
                histogram_value_unit = 1 / (function_dict[axis_unit] ** 2)
            else:
                histogram_value_unit = u.dimensionless_unscaled
            hist_1d_list = hist_1d_list * histogram_value_unit
            for i_histogram, _ in enumerate(x_bin_edges_list):
                if self.individual_telescopes:
                    hdf5_table_name = (
                        f"/{function_dict['file name']}_"
                        f"tel_index_{self.telescope_indices[i_histogram]}"
                    )
                else:
                    hdf5_table_name = f"/{function_dict['file name']}_all_tels"

                table = fill_hdf5_table(
                    hist=hist_1d_list[i_histogram],
                    x_bin_edges=x_bin_edges_list[i_histogram],
                    y_bin_edges=None,
                    x_label=function_dict["bin edges"],
                    y_label=None,
                    meta_data=self._meta_dict,
                )
                self._logger.info(
                    f"Writing 1D histogram with name {hdf5_table_name} to {self.hdf5_file_name}."
                )
                # overwrite takes precedence over append
                if overwrite is True:
                    append = False
                else:
                    append = True
                write_table(
                    table, self.hdf5_file_name, hdf5_table_name, append=append, overwrite=overwrite
                )

    @property
    def dict_2d_distributions(self):
        """
        Dictionary to label the 2D distributions according to the class methods.

        Returns
        -------
        dict:
            The dictionary with information about the 2D distributions.
        """
        fn_key = "function"
        file_name = "file name"
        title = "title"
        x_bin_edges = "x bin edges"
        x_axis_unit = "x axis unit"
        y_bin_edges = "y bin edges"
        y_axis_unit = "y axis unit"
        if self._dict_2d_distributions is None:
            self._dict_2d_distributions = {
                "counts": {
                    fn_key: "get_2d_photon_position_distr",
                    file_name: "hist_2d_photon_count_distr",
                    title: "Photon count distribution on the ground",
                    x_bin_edges: "x position on the ground",
                    x_axis_unit: self.hist_config["hist_position"][X_AXIS_STRING]["start"].unit,
                    y_bin_edges: "y position on the ground",
                    y_axis_unit: self.hist_config["hist_position"][Y_AXIS_STRING]["start"].unit,
                },
                "density": {
                    fn_key: "get_2d_photon_density_distr",
                    file_name: "hist_2d_photon_density_distr",
                    title: "Photon density distribution on the ground",
                    x_bin_edges: "x position on the ground",
                    x_axis_unit: self.hist_config["hist_position"][X_AXIS_STRING]["start"].unit,
                    y_bin_edges: "y position on the ground",
                    y_axis_unit: self.hist_config["hist_position"][Y_AXIS_STRING]["start"].unit,
                },
                "direction": {
                    fn_key: "get_2d_photon_direction_distr",
                    file_name: "hist_2d_photon_direction_distr",
                    title: "Photon arrival direction",
                    x_bin_edges: "x direction cosine",
                    x_axis_unit: u.dimensionless_unscaled,
                    y_bin_edges: "y direction cosine",
                    y_axis_unit: u.dimensionless_unscaled,
                },
                "time_altitude": {
                    fn_key: "get_2d_photon_time_altitude_distr",
                    file_name: "hist_2d_photon_time_altitude_distr",
                    title: "Time of arrival vs altitude of emission",
                    x_bin_edges: "Time of arrival",
                    x_axis_unit: self.hist_config["hist_time_altitude"][X_AXIS_STRING][
                        "start"
                    ].unit,
                    y_bin_edges: "Altitude of emission",
                    y_axis_unit: self.hist_config["hist_time_altitude"][Y_AXIS_STRING][
                        "start"
                    ].unit,
                },
                "num_photons_per_telescope": {
                    fn_key: "get_2d_num_photons_distr",
                    file_name: "hist_2d_photon_telescope_event_distr",
                    title: "Number of photons per telescope and per event",
                    x_bin_edges: "Telescope counter",
                    x_axis_unit: u.dimensionless_unscaled,
                    y_bin_edges: "Event counter",
                    y_axis_unit: u.dimensionless_unscaled,
                },
            }
        return self._dict_2d_distributions

    def _export_2d_histograms(self, overwrite):
        """
        Auxiliary function to export only the 2D histograms.

        Parameters
        ----------
        overwrite: bool
            If True overwrites the histograms already saved in the hdf5 file.
        """
        x_axis_unit = "x axis unit"
        y_axis_unit = "y axis unit"

        for property_name, function_dict in self.dict_2d_distributions.items():
            self._meta_dict["Title"] = sanitize_name(function_dict["title"])
            histogram_function = getattr(self, function_dict["function"])

            hist_2d_list, x_bin_edges_list, y_bin_edges_list = histogram_function()
            if function_dict["function"] == "get_2d_photon_density_distr":
                histogram_value_unit = 1 / (
                    self.dict_2d_distributions[property_name][x_axis_unit]
                    * self.dict_2d_distributions[property_name][y_axis_unit]
                )
            else:
                histogram_value_unit = u.dimensionless_unscaled

            hist_2d_list, x_bin_edges_list, y_bin_edges_list = (
                hist_2d_list * histogram_value_unit,
                x_bin_edges_list * self.dict_2d_distributions[property_name][x_axis_unit],
                y_bin_edges_list * self.dict_2d_distributions[property_name][y_axis_unit],
            )

            for i_histogram, _ in enumerate(x_bin_edges_list):
                if self.individual_telescopes:
                    hdf5_table_name = (
                        f"/{self.dict_2d_distributions[property_name]['file name']}"
                        f"_tel_index_{self.telescope_indices[i_histogram]}"
                    )
                else:
                    hdf5_table_name = (
                        f"/{self.dict_2d_distributions[property_name]['file name']}_all_tels"
                    )
                table = fill_hdf5_table(
                    hist=hist_2d_list[i_histogram],
                    x_bin_edges=x_bin_edges_list[i_histogram],
                    y_bin_edges=y_bin_edges_list[i_histogram],
                    x_label=function_dict["x bin edges"],
                    y_label=function_dict["y bin edges"],
                    meta_data=self._meta_dict,
                )

                self._logger.info(
                    f"Writing 2D histogram with name {hdf5_table_name} to {self.hdf5_file_name}."
                )
                # Always appending to table due to the file previously created
                # by self._export_1d_histograms.
                write_table(
                    table, self.hdf5_file_name, hdf5_table_name, append=True, overwrite=overwrite
                )

    def export_event_header_1d_histogram(
        self, event_header_element, bins=50, hist_range=None, overwrite=False
    ):
        """
        Export 'event_header_element' from CORSIKA to hd5 for a 1D histogram.

        Parameters
        ----------
        event_header_element: str
            The key to the CORSIKA event header element.
            Possible choices are stored in 'self.all_event_keys'.
        bins: float
            Number of bins for the histogram.
        hist_range: 2-tuple
            Tuple to define the range of the histogram.
        overwrite: bool
            If True overwrites the histograms already saved in the hdf5 file.
        """
        hist, bin_edges = self.event_1d_histogram(
            event_header_element, bins=bins, hist_range=hist_range
        )
        bin_edges *= self.event_information[event_header_element].unit
        table = fill_hdf5_table(
            hist=hist,
            x_bin_edges=bin_edges,
            y_bin_edges=None,
            x_label=event_header_element,
            y_label=None,
            meta_data=self._meta_dict,
        )
        hdf5_table_name = f"/event_2d_histograms_{event_header_element}"

        self._logger.info(
            f"Exporting histogram with name {hdf5_table_name} to {self.hdf5_file_name}."
        )
        # overwrite takes precedence over append
        if overwrite is True:
            append = False
        else:
            append = True
        write_table(table, self.hdf5_file_name, hdf5_table_name, append=append, overwrite=overwrite)

    def export_event_header_2d_histogram(
        self,
        event_header_element_1,
        event_header_element_2,
        bins=50,
        hist_range=None,
        overwrite=False,
    ):
        """
        Export event_header of a 2D histogram to a hdf5 file.

        Searches the 2D histogram for the key 'event_header_element_1' and
        'event_header_element_2'from the CORSIKA event header.

        Parameters
        ----------
        event_header_element_1: str
            The key to the CORSIKA event header element.
        event_header_element_2: str
            The key to the CORSIKA event header element.
            Possible choices for 'event_header_element_1' and 'event_header_element_2' are stored
            in 'self.all_event_keys'.
        bins: float
            Number of bins for the histogram.
        hist_range: 2-tuple
            Tuple to define the range of the histogram.
        overwrite: bool
            If True overwrites the histograms already saved in the hdf5 file.

        """
        hist, x_bin_edges, y_bin_edges = self.event_2d_histogram(
            event_header_element_1, event_header_element_2, bins=bins, hist_range=hist_range
        )
        x_bin_edges *= self.event_information[event_header_element_1].unit
        y_bin_edges *= self.event_information[event_header_element_2].unit

        table = fill_hdf5_table(
            hist=hist,
            x_bin_edges=x_bin_edges,
            y_bin_edges=y_bin_edges,
            x_label=event_header_element_1,
            y_label=event_header_element_2,
            meta_data=self._meta_dict,
        )

        hdf5_table_name = f"/event_2d_histograms_{event_header_element_1}_{event_header_element_2}"

        self._logger.info(
            f"Exporting histogram with name {hdf5_table_name} to {self.hdf5_file_name}."
        )
        # overwrite takes precedence over append
        if overwrite is True:
            append = False
        else:
            append = True
        write_table(table, self.hdf5_file_name, hdf5_table_name, append=append, overwrite=overwrite)

    @property
    def num_photons_per_telescope(self):
        """
        The number of photons per event, considering the telescopes given by self.telescope_indices.

        Returns
        -------
        numpy.array
            Number of photons per telescope.
        """
        self._num_photons_per_telescope = np.sum(
            np.array(self.num_photons_per_event_per_telescope), axis=1
        )
        return self._num_photons_per_telescope

    @property
    def total_num_photons(self):
        """
        The total number of photons.

        Returns
        -------
        float
            Total number photons.
        """
        self._total_num_photons = np.sum(self.num_photons_per_event)
        return self._total_num_photons

    @property
    def telescope_positions(self):
        """
        The telescope positions found in the CORSIKA output file.

        It does not depend on the telescope_indices attribute.

        Returns
        -------
        numpy.ndarray
            x, y and z positions of the telescopes and their radius according to the CORSIKA
            spherical representation of the telescopes.
        """
        return self._telescope_positions

    @telescope_positions.setter
    def telescope_positions(self, new_positions):
        """
        Set the telescope positions.

        Parameters
        ----------
        numpy.ndarray
            x, y and z positions of the telescopes and their radius according to the CORSIKA
            spherical representation of the telescopes.
        """
        self._telescope_positions = new_positions

    # In the next five functions, we provide dedicated functions to retrieve specific information
    # about the runs, i.e. zenith, azimuth, total energy, interaction height and Earth magnetic
    # field defined for the run. For other information, please use the get_event_parameter_info
    # function.
    @property
    def event_zenith_angles(self):
        """
        Get the zenith angles of the simulated events in astropy units of degrees.

        Returns
        -------
        astropy.Quantity
            The zenith angles for each event.
        """
        if self._event_zenith_angles is None:
            self._event_zenith_angles = np.around(
                (self.event_information["zenith"]).to(u.deg),
                4,
            )
        return self._event_zenith_angles

    @property
    def event_azimuth_angles(self):
        """
        Get the azimuth angles of the simulated events in astropy units of degrees.

        Returns
        -------
        astropy.Quantity
            The azimuth angles for each event, usually in degrees.
        """
        if self._event_azimuth_angles is None:
            self._event_azimuth_angles = np.around(
                (self.event_information["azimuth"]).to(u.deg),
                4,
            )
        return self._event_azimuth_angles

    @property
    def event_energies(self):
        """
        Get the energy of the simulated events in astropy units of TeV.

        Returns
        -------
        astropy.Quantity
            The total energies of the incoming particles for each event, usually in TeV.
        """
        if self._event_total_energies is None:
            self._event_total_energies = np.around(
                (self.event_information["total_energy"]).to(u.TeV),
                4,
            )
        return self._event_total_energies

    @property
    def event_first_interaction_heights(self):
        """
        Get the height of the first interaction in astropy units of km.

        If negative, tracking starts at margin of atmosphere,
        see TSTART in the CORSIKA 7 user guide.

        Returns
        -------
        astropy.Quantity
            The first interaction height for each event, usually in km.
        """
        if self._event_first_interaction_heights is None:
            self._event_first_interaction_heights = np.around(
                (self.event_information["first_interaction_height"]).to(u.km),
                4,
            )
        return self._event_first_interaction_heights

    @property
    def magnetic_field(self):
        """
        Get the Earth magnetic field from the events header in astropy units of microT.

        A tuple with Earth's magnetic field in the x and z directions are returned.

        Returns
        -------
        astropy.Quantity
            The Earth magnetic field in the x direction used for each event.
        astropy.Quantity
            The Earth magnetic field in the z direction used for each event.
        """
        if self._magnetic_field_x is None:
            self._magnetic_field_x = (self.event_information["earth_magnetic_field_x"]).to(u.uT)
        if self._magnetic_field_z is None:
            self._magnetic_field_z = (self.event_information["earth_magnetic_field_z"]).to(u.uT)
        return self._magnetic_field_x, self._magnetic_field_z

    def get_event_parameter_info(self, parameter):
        """
        Get specific information (i.e. any parameter) of the events.

        The parameter is passed through the key word parameter.
        Available options are to be found under self.all_event_keys.
        The unit of the parameter, if any, is given according to the CORSIKA version
        (please see user guide in this case).

        Parameters
        ----------
        parameter: str
            The parameter of interest. Available options are to be found under
            self.all_event_keys.

        Returns
        -------
        astropy.Quantity
            The array with the event information as required by the parameter.

        Raises
        ------
        KeyError:
            If parameter is not valid.
        """
        if parameter not in self.all_event_keys:
            msg = f"key is not valid. Valid entries are {self.all_event_keys}"
            self._logger.error(msg)
            raise KeyError
        return self.event_information[parameter]

    def get_run_info(self, parameter):
        """
        Get specific information (i.e. any parameter) of the run.

        The parameter is passed through the key word parameter.
        Available options are to be found under self.all_run_keys.
        The unit of the parameter, if any, is given according to the CORSIKA version
        (please see user guide in this case).

        Parameters
        ----------
        parameter: str
            The parameter of interest. Available options are to be found under
            self.all_run_keys.

        Raises
        ------
        KeyError:
            If parameter is not valid.
        """
        if parameter not in self.all_run_keys:
            msg = f"key is not valid. Valid entries are {self.all_run_keys}"
            self._logger.error(msg)
            raise KeyError
        return self.header[parameter]

    def event_1d_histogram(self, key, bins=50, hist_range=None):
        """
        Create a histogram for the all events using key as parameter.

        Valid keys are stored in self.all_event_keys (CORSIKA defined).

        Parameters
        ----------
        key: str
            The information from which to build the histogram, e.g. total_energy, zenith or
            first_interaction_height.
        bins: float
            Number of bins for the histogram.
        hist_range: 2-tuple
            Tuple to define the range of the histogram.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            Edges of the histogram.


        Raises
        ------
        KeyError:
            If key is not valid.
        """
        if key not in self.all_event_keys:
            msg = f"key is not valid. Valid entries are {self.all_event_keys}"
            self._logger.error(msg)
            raise KeyError
        hist, bin_edges = np.histogram(
            self.event_information[key].value,
            bins=bins,
            range=hist_range,
        )
        return hist, bin_edges

    def event_2d_histogram(self, key_1, key_2, bins=50, hist_range=None):
        """
        Create a 2D histogram for the all events using key_1 and key_2 as parameters.

        Valid keys are stored in self.all_event_keys (CORSIKA defined).

        Parameters
        ----------
        key_1: str
            The information from which to build the histogram, e.g. total_energy, zenith or
            first_interaction_height.
        key_2: str
            The information from which to build the histogram, e.g. total_energy, zenith or
            first_interaction_height.
        bins: float
            Number of bins for the histogram.
        hist_range: 2-tuple
            Tuple to define the range of the histogram.

        Returns
        -------
        numpy.ndarray
            The counts of the histogram.
        numpy.array
            x Edges of the histogram.
        numpy.array
            y Edges of the histogram.


        Raises
        ------
        KeyError:
            If at least one of the keys is not valid.
        """
        for key in [key_1, key_2]:
            if key not in self.all_event_keys:
                msg = (
                    f"At least one of the keys given is not valid. Valid entries are "
                    f"{self.all_event_keys}"
                )
                self._logger.error(msg)
                raise KeyError
        hist, x_bin_edges, y_bin_edges = np.histogram2d(
            self.event_information[key_1].value,
            self.event_information[key_2].value,
            bins=bins,
            range=hist_range,
        )
        return hist, x_bin_edges, y_bin_edges



"""
PSF parameter optimisation and fitting routines for mirror alignment and reflection parameters.

This module provides functions for loading PSF data, generating random parameter sets,
running PSF simulations, calculating RMSD, and finding the best-fit parameters for a given
telescope model.
PSF (Point Spread Function) describes how a point source of light is spread out by the
optical system, and RMSD (Root Mean Squared Deviation) is used as the optimization metric
to quantify the difference between measured and simulated PSF curves.
"""

import logging
from collections import OrderedDict

import astropy.units as u
import numpy as np
from astropy.table import Table
from scipy import stats

from simtools.data_model import model_data_writer as writer
from simtools.ray_tracing.ray_tracing import RayTracing
from simtools.utils import general as gen
from simtools.visualization import plot_psf
from simtools.visualization.plot_psf import DEFAULT_FRACTION, get_psf_diameter_label

logger = logging.getLogger(__name__)


# Constants
RADIUS = "Radius"
CUMULATIVE_PSF = "Cumulative PSF"
KS_STATISTIC_NAME = "KS statistic"


def _create_log_header_and_format_value(title, tel_model, additional_info=None, value=None):
    """Create log header and format parameter values."""
    if value is not None:  # Format value mode
        if isinstance(value, list):
            return "[" + ", ".join([f"{v:.6f}" for v in value]) + "]"
        if isinstance(value, int | float):
            return f"{value:.6f}"
        return str(value)

    # Create header mode
    header_lines = [f"# {title}", f"# Telescope: {tel_model.name}"]
    if additional_info:
        for key, val in additional_info.items():
            header_lines.append(f"# {key}: {val}")
    header_lines.extend(["#" + "=" * 65, ""])
    return "\n".join(header_lines) + "\n"


def calculate_rmsd(data, sim):
    """Calculate RMSD between measured and simulated cumulative PSF curves."""
    return np.sqrt(np.mean((data - sim) ** 2))


def calculate_ks_statistic(data, sim):
    """Calculate the KS statistic between measured and simulated cumulative PSF curves."""
    return stats.ks_2samp(data, sim)


def get_previous_values(tel_model):
    """
    Retrieve current PSF parameter values from the telescope model.

    Parameters
    ----------
    tel_model : TelescopeModel
        Telescope model object containing parameter configurations.

    Returns
    -------
    dict
        Dictionary containing current values of PSF optimization parameters:
        - 'mirror_reflection_random_angle': Random reflection angle parameters
        - 'mirror_align_random_horizontal': Horizontal alignment parameters
        - 'mirror_align_random_vertical': Vertical alignment parameters
    """
    return {
        "mirror_reflection_random_angle": tel_model.get_parameter_value(
            "mirror_reflection_random_angle"
        ),
        "mirror_align_random_horizontal": tel_model.get_parameter_value(
            "mirror_align_random_horizontal"
        ),
        "mirror_align_random_vertical": tel_model.get_parameter_value(
            "mirror_align_random_vertical"
        ),
    }


def _run_ray_tracing_simulation(tel_model, site_model, args_dict, pars):
    """Run a ray tracing simulation with the given telescope parameters."""
    if pars is None:
        raise ValueError("No best parameters found")

    tel_model.overwrite_parameters(pars)
    ray = RayTracing(
        telescope_model=tel_model,
        site_model=site_model,
        simtel_path=args_dict["simtel_path"],
        zenith_angle=args_dict["zenith"] * u.deg,
        source_distance=args_dict["src_distance"] * u.km,
        off_axis_angle=[0.0] * u.deg,
    )
    ray.simulate(test=args_dict.get("test", False), force=True)
    ray.analyze(force=True, use_rx=False)
    im = ray.images()[0]
    fraction = args_dict.get("fraction", DEFAULT_FRACTION)
    return im.get_psf(fraction=fraction), im


def run_psf_simulation(
    tel_model,
    site,
    args_dict,
    pars,
    data_to_plot,
    radius,
    pdf_pages=None,
    is_best=False,
    use_ks_statistic=False,
):
    """
    Run PSF simulation for given parameters and calculate optimization metric.

    Parameters
    ----------
    tel_model : TelescopeModel
        Telescope model object to be configured with the test parameters.
    site : Site
        Site model object with environmental conditions.
    args_dict : dict
        Dictionary containing simulation configuration arguments.
    pars : dict
        Dictionary of parameter values to test in the simulation.
    data_to_plot : dict
        Dictionary containing measured PSF data under "measured" key.
    radius : array-like
        Radius values in cm for PSF evaluation.
    pdf_pages : PdfPages, optional
        PDF pages object for saving plots (default: None).
    is_best : bool, optional
        Flag indicating if this is the best parameter set (default: False).
    use_ks_statistic : bool, optional
        If True, use KS statistic as metric; if False, use RMSD (default: False).

    Returns
    -------
    tuple of (float, float, float or None, array)
        - psf_diameter: PSF containment diameter of the simulated PSF in cm
        - metric: RMSD or KS statistic value
        - p_value: p-value from KS test (None if using RMSD)
        - simulated_data: Structured array with simulated cumulative PSF data
    """
    psf_diameter, im = _run_ray_tracing_simulation(tel_model, site, args_dict, pars)

    if radius is None:
        raise ValueError("Radius data is not available.")

    simulated_data = im.get_cumulative_data(radius * u.cm)

    if use_ks_statistic:
        ks_statistic, p_value = calculate_ks_statistic(
            data_to_plot["measured"][CUMULATIVE_PSF], simulated_data[CUMULATIVE_PSF]
        )
        metric = ks_statistic
    else:
        metric = calculate_rmsd(
            data_to_plot["measured"][CUMULATIVE_PSF], simulated_data[CUMULATIVE_PSF]
        )
        p_value = None

    # Handle plotting if requested
    if pdf_pages is not None and args_dict.get("plot_all", False):
        data_to_plot["simulated"] = simulated_data
        plot_psf.create_psf_parameter_plot(
            data_to_plot,
            pars,
            psf_diameter,
            metric,
            is_best,
            pdf_pages,
            fraction=args_dict.get("fraction", DEFAULT_FRACTION),
            p_value=p_value,
            use_ks_statistic=use_ks_statistic,
        )
        del data_to_plot["simulated"]

    return psf_diameter, metric, p_value, simulated_data


def load_and_process_data(args_dict):
    """
    Load and process PSF measurement data from ECSV file.

    Parameters
    ----------
    args_dict : dict
        Dictionary containing command-line arguments with 'data' and 'model_path' keys.

    Returns
    -------
    tuple of (OrderedDict, array)
        - data_dict: OrderedDict with "measured" key containing structured array
          of radius and cumulative PSF data
        - radius: Array of radius values in cm

    Raises
    ------
    FileNotFoundError
        If no data file is specified in args_dict.
    """
    if args_dict["data"] is None:
        raise FileNotFoundError("No data file specified for PSF optimization.")

    data_file = gen.find_file(args_dict["data"], args_dict["model_path"])
    table = Table.read(data_file, format="ascii.ecsv")

    radius_column = next((col for col in table.colnames if "radius" in col.lower()), None)
    integral_psf_column = next((col for col in table.colnames if "integral" in col.lower()), None)

    # Create structured array with converted data
    d_type = {"names": (RADIUS, CUMULATIVE_PSF), "formats": ("f8", "f8")}
    data = np.zeros(len(table), dtype=d_type)

    data[RADIUS] = table[radius_column].to(u.cm).value
    data[CUMULATIVE_PSF] = table[integral_psf_column]
    data[CUMULATIVE_PSF] /= np.max(np.abs(data[CUMULATIVE_PSF]))  # Normalize to max = 1.0

    return OrderedDict([("measured", data)]), data[RADIUS]


def write_tested_parameters_to_file(
    results, best_pars, best_psf_diameter, output_dir, tel_model, fraction=DEFAULT_FRACTION
):
    """
    Write optimization results and tested parameters to a log file.

    Parameters
    ----------
    results : list
        List of tuples containing (parameters, ks_statistic, p_value, psf_diameter, simulated_data)
        for each tested parameter set.
    best_pars : dict
        Dictionary containing the best parameter values found.
    best_psf_diameter : float
        PSF containment diameter in cm for the best parameter set.
    output_dir : Path
        Directory where the log file will be written.
    tel_model : TelescopeModel
        Telescope model object for naming the output file.
    fraction : float, optional
        PSF containment fraction for labeling (default: 0.8).

    Returns
    -------
    Path
        Path to the created log file.
    """
    param_file = output_dir.joinpath(f"psf_optimization_{tel_model.name}.log")
    psf_label = get_psf_diameter_label(fraction)

    with open(param_file, "w", encoding="utf-8") as f:
        header = _create_log_header_and_format_value(
            "PSF Parameter Optimization Log",
            tel_model,
            {"Total parameter sets tested": len(results)},
        )
        f.write(header)

        f.write("PARAMETER TESTING RESULTS:\n")
        for i, (pars, ks_statistic, p_value, psf_diameter, _) in enumerate(results):
            status = "BEST" if pars is best_pars else "TESTED"
            f.write(
                f"[{status}] Set {i + 1:03d}: KS_stat={ks_statistic:.5f}, "
                f"p_value={p_value:.5f}, {psf_label}={psf_diameter:.5f} cm\n"
            )
            for par, value in pars.items():
                f.write(f"    {par}: {value}\n")
            f.write("\n")

        f.write("OPTIMIZATION SUMMARY:\n")
        f.write(f"Best KS statistic: {min(result[1] for result in results):.5f}\n")
        f.write(f"Best {psf_label}: {best_psf_diameter:.5f} cm\n")
        f.write("\nOPTIMIZED PARAMETERS:\n")
        for par, value in best_pars.items():
            f.write(f"{par}: {value}\n")
    return param_file


def _add_units_to_psf_parameters(best_pars):
    """Add astropy units to PSF parameters based on their schemas."""
    psf_pars_with_units = {}
    for param_name, param_values in best_pars.items():
        if param_name == "mirror_reflection_random_angle":
            psf_pars_with_units[param_name] = [
                param_values[0] * u.deg,
                param_values[1] * u.dimensionless_unscaled,
                param_values[2] * u.deg,
            ]
        elif param_name in ["mirror_align_random_horizontal", "mirror_align_random_vertical"]:
            psf_pars_with_units[param_name] = [
                param_values[0] * u.deg,
                param_values[1] * u.deg,
                param_values[2] * u.dimensionless_unscaled,
                param_values[3] * u.dimensionless_unscaled,
            ]
        else:
            psf_pars_with_units[param_name] = param_values
    return psf_pars_with_units


def export_psf_parameters(best_pars, telescope, parameter_version, output_dir):
    """
    Export optimized PSF parameters as simulation model parameter files.

    Parameters
    ----------
    best_pars : dict
        Dictionary containing the optimized parameter values.
    telescope : str
        Telescope name for the parameter files.
    parameter_version : str
        Version string for the parameter files.
    output_dir : Path
        Base directory for parameter file output.

    Notes
    -----
    Creates individual JSON files for each optimized parameter with
    units. Files are saved in the format:
    {output_dir}/{telescope}/{parameter_name}-{parameter_version}.json

    Raises
    ------
    ValueError, KeyError, OSError
        If parameter export fails due to invalid values, missing keys, or file I/O errors.
    """
    try:
        psf_pars_with_units = _add_units_to_psf_parameters(best_pars)
        parameter_output_path = output_dir.parent / telescope
        for parameter_name, parameter_value in psf_pars_with_units.items():
            writer.ModelDataWriter.dump_model_parameter(
                parameter_name=parameter_name,
                value=parameter_value,
                instrument=telescope,
                parameter_version=parameter_version,
                output_file=f"{parameter_name}-{parameter_version}.json",
                output_path=parameter_output_path,
            )
        logger.info(f"simulation model parameter files exported to {output_dir}")

    except (ValueError, KeyError, OSError) as e:
        logger.error(f"Error exporting simulation parameters: {e}")


def _calculate_param_gradient(
    tel_model,
    site_model,
    args_dict,
    current_params,
    data_to_plot,
    radius,
    current_rmsd,
    param_name,
    param_values,
    epsilon,
    use_ks_statistic,
):
    """
    Calculate numerical gradient for a single parameter using finite differences.

    The gradient is calculated using forward finite differences:
    gradient = (f(x + epsilon) - f(x)) / epsilon

    Parameters
    ----------
    tel_model : TelescopeModel
        The telescope model object containing the current parameter configuration.
    site_model : SiteModel
        The site model object with environmental conditions.
    args_dict : dict
        Dictionary containing simulation arguments and configuration options.
    current_params : dict
        Dictionary of current parameter values for all optimization parameters.
    data_to_plot : dict
        Dictionary containing measured PSF data with "measured" key.
    radius : array-like
        Radius values in cm for PSF evaluation.
    current_rmsd : float
        Current RMSD at the current parameter configuration.
    param_name : str
        Name of the parameter for which to calculate the gradient.
    param_values : float or list
        Current value(s) of the parameter. Can be a single value or list of values.
    epsilon : float
        Small perturbation value for finite difference calculation.
    use_ks_statistic : bool
        If True, calculate gradient with respect to KS statistic; if False, use RMSD.

    Returns
    -------
    float or list
        Gradient value(s) for the parameter. Returns a single float if param_values
        is a single value, or a list of gradients if param_values is a list.

    If a simulation fails during gradient calculation, a gradient of 0.0 is assigned
    for that component to ensure the optimization can continue.
    """
    param_gradients = []
    values_list = param_values if isinstance(param_values, list) else [param_values]

    for i, value in enumerate(values_list):
        perturbed_params = {
            k: v.copy() if isinstance(v, list) else v for k, v in current_params.items()
        }

        if isinstance(param_values, list):
            perturbed_params[param_name][i] = value + epsilon
        else:
            perturbed_params[param_name] = value + epsilon

        try:
            _, perturbed_rmsd, _, _ = run_psf_simulation(
                tel_model,
                site_model,
                args_dict,
                perturbed_params,
                data_to_plot,
                radius,
                pdf_pages=None,
                is_best=False,
                use_ks_statistic=use_ks_statistic,
            )
            param_gradients.append((perturbed_rmsd - current_rmsd) / epsilon)
        except (ValueError, RuntimeError):
            param_gradients.append(0.0)

    return param_gradients[0] if not isinstance(param_values, list) else param_gradients


def calculate_gradient(
    tel_model,
    site_model,
    args_dict,
    current_params,
    data_to_plot,
    radius,
    current_rmsd,
    epsilon=0.0005,
    use_ks_statistic=False,
):
    """
    Calculate numerical gradients for all optimization parameters.

    Parameters
    ----------
    tel_model : TelescopeModel
        Telescope model object for simulations.
    site_model : SiteModel
        Site model object with environmental conditions.
    args_dict : dict
        Dictionary containing simulation configuration arguments.
    current_params : dict
        Dictionary of current parameter values for all optimization parameters.
    data_to_plot : dict
        Dictionary containing measured PSF data.
    radius : array-like
        Radius values in cm for PSF evaluation.
    current_rmsd : float
        Current RMSD or KS statistic value.
    epsilon : float, optional
        Perturbation value for finite difference calculation (default: 0.0005).
    use_ks_statistic : bool, optional
        If True, calculate gradients for KS statistic; if False, use RMSD (default: False).

    Returns
    -------
    dict
        Dictionary mapping parameter names to their gradient values.
        For parameters with multiple components, gradients are returned as lists.
    """
    gradients = {}

    for param_name, param_values in current_params.items():
        gradients[param_name] = _calculate_param_gradient(
            tel_model,
            site_model,
            args_dict,
            current_params,
            data_to_plot,
            radius,
            current_rmsd,
            param_name,
            param_values,
            epsilon,
            use_ks_statistic,
        )

    return gradients


def apply_gradient_step(current_params, gradients, learning_rate):
    """
    Apply gradient descent step to update parameters.

    Parameters
    ----------
    current_params : dict
        Dictionary of current parameter values.
    gradients : dict
        Dictionary of gradient values for each parameter.
    learning_rate : float
        Step size for the gradient descent update.

    Returns
    -------
    dict
        Dictionary of updated parameter values after applying the gradient step.
    """
    new_params = {}
    for param_name, param_values in current_params.items():
        param_gradients = gradients[param_name]

        if isinstance(param_values, list):
            new_params[param_name] = [
                value - learning_rate * gradient
                for value, gradient in zip(param_values, param_gradients)
            ]
        else:
            new_params[param_name] = param_values - learning_rate * param_gradients

    return new_params


def _perform_gradient_step_with_retries(
    tel_model,
    site_model,
    args_dict,
    current_params,
    current_metric,
    data_to_plot,
    radius,
    learning_rate,
    max_retries=3,
):
    """
    Attempt gradient descent step with adaptive learning rate reduction on rejection.

    The learning rate reduction strategy follows these rules:
    - If step is rejected: learning_rate *= 0.7
    - If attempt number < number of max retries then try again
    - If learning_rate drops below 1e-5: reset to 0.001
    - If all retries fail: returns None values with step_accepted=False

    This adaptive approach helps navigate local minima and ensures robust convergence
    by automatically adjusting the step size based on optimization progress.

    Parameters
    ----------
    tel_model : TelescopeModel
        Telescope model object containing the current parameter configuration.
    site_model : SiteModel
        Site model object with environmental conditions for ray tracing simulations.
    args_dict : dict
        Dictionary containing simulation configuration arguments and settings.
    current_params : dict
        Dictionary of current parameter values for all optimization parameters.
    current_metric : float
        Current optimization metric value (RMSD or KS statistic) to improve upon.
    data_to_plot : dict
        Dictionary containing measured PSF data under "measured" key for comparison.
    radius : array-like
        Radius values in cm for PSF evaluation and comparison.
    learning_rate : float
        Initial learning rate for the gradient descent step.
    max_retries : int, optional
        Maximum number of attempts with learning rate reduction (default: 3).

    Returns
    -------
    tuple of (dict, float, float, float or None, array, bool, float)
        - new_params: Updated parameter dictionary if step accepted, None if rejected
        - new_psf_diameter: PSF containment diameter in cm for new parameters, None if step rejected
        - new_metric: New optimization metric value, None if step rejected
        - new_p_value: p-value from KS test if applicable, None otherwise
        - new_simulated_data: Simulated PSF data array, None if step rejected
        - step_accepted: Boolean indicating if any step was accepted
        - final_learning_rate: Learning rate after potential reductions

    """
    current_lr = learning_rate

    for attempt in range(max_retries):
        try:
            gradients = calculate_gradient(
                tel_model,
                site_model,
                args_dict,
                current_params,
                data_to_plot,
                radius,
                current_metric,
                use_ks_statistic=False,
            )
            new_params = apply_gradient_step(current_params, gradients, current_lr)

            new_psf_diameter, new_metric, new_p_value, new_simulated_data = run_psf_simulation(
                tel_model,
                site_model,
                args_dict,
                new_params,
                data_to_plot,
                radius,
                pdf_pages=None,
                is_best=False,
                use_ks_statistic=False,
            )

            if new_metric < current_metric:
                return (
                    new_params,
                    new_psf_diameter,
                    new_metric,
                    new_p_value,
                    new_simulated_data,
                    True,
                    current_lr,
                )

            logger.info(
                f"Step rejected (RMSD {current_metric:.6f} -> {new_metric:.6f}), "
                f"reducing learning rate {current_lr:.6f} -> {current_lr * 0.7:.6f}"
            )
            current_lr *= 0.7

            if current_lr < 1e-5:
                current_lr = 0.001

        except (ValueError, RuntimeError, KeyError) as e:
            logger.warning(f"Simulation failed on attempt {attempt + 1}: {e}")
            continue

    return None, None, None, None, None, False, current_lr


def _create_step_plot(
    pdf_pages,
    args_dict,
    data_to_plot,
    current_params,
    new_psf_diameter,
    new_metric,
    new_p_value,
    new_simulated_data,
):
    """Create plot for an accepted gradient step."""
    if pdf_pages is None or not args_dict.get("plot_all", False) or new_simulated_data is None:
        return

    data_to_plot["simulated"] = new_simulated_data
    plot_psf.create_psf_parameter_plot(
        data_to_plot,
        current_params,
        new_psf_diameter,
        new_metric,
        False,
        pdf_pages,
        fraction=args_dict.get("fraction", DEFAULT_FRACTION),
        p_value=new_p_value,
        use_ks_statistic=False,
    )
    del data_to_plot["simulated"]


def _create_final_plot(
    pdf_pages,
    tel_model,
    site_model,
    args_dict,
    best_params,
    data_to_plot,
    radius,
    best_psf_diameter,
):
    """Create final plot for best parameters."""
    if pdf_pages is None or best_params is None:
        return

    logger.info("Creating final plot for best parameters with both RMSD and KS statistic...")
    _, best_ks_stat, best_p_value, best_simulated_data = run_psf_simulation(
        tel_model,
        site_model,
        args_dict,
        best_params,
        data_to_plot,
        radius,
        pdf_pages=None,
        is_best=False,
        use_ks_statistic=True,
    )
    best_rmsd = calculate_rmsd(
        data_to_plot["measured"][CUMULATIVE_PSF], best_simulated_data[CUMULATIVE_PSF]
    )

    data_to_plot["simulated"] = best_simulated_data
    plot_psf.create_psf_parameter_plot(
        data_to_plot,
        best_params,
        best_psf_diameter,
        best_rmsd,
        True,
        pdf_pages,
        fraction=args_dict.get("fraction", DEFAULT_FRACTION),
        p_value=best_p_value,
        use_ks_statistic=False,
        second_metric=best_ks_stat,
    )
    del data_to_plot["simulated"]
    pdf_pages.close()
    logger.info("Cumulative PSF plots saved")


def run_gradient_descent_optimization(
    tel_model,
    site_model,
    args_dict,
    data_to_plot,
    radius,
    rmsd_threshold,
    learning_rate,
    output_dir,
):
    """
    Run gradient descent optimization to minimize PSF fitting metric.

    Parameters
    ----------
    tel_model : TelescopeModel
        Telescope model object to be optimized.
    site_model : SiteModel
        Site model object with environmental conditions.
    args_dict : dict
        Dictionary containing simulation configuration arguments.
    data_to_plot : dict
        Dictionary containing measured PSF data under "measured" key.
    radius : array-like
        Radius values in cm for PSF evaluation.
    rmsd_threshold : float
        Convergence threshold for RMSD improvement.
    learning_rate : float
        Initial learning rate for gradient descent steps.
    output_dir : Path
        Directory for saving optimization plots and results.

    Returns
    -------
    tuple of (dict, float, list)
        - best_params: Dictionary of optimized parameter values
        - best_psf_diameter: PSF containment diameter in cm for the best parameters
        - results: List of (params, metric, p_value, psf_diameter, simulated_data)
          for each iteration

    Returns None values if optimization fails or no measurement data is provided.
    """
    if data_to_plot is None or radius is None:
        logger.error("No PSF measurement data provided. Cannot run optimization.")
        return None, None, []

    current_params = get_previous_values(tel_model)
    pdf_pages = plot_psf.setup_pdf_plotting(args_dict, output_dir, tel_model.name)
    results = []

    # Evaluate initial parameters
    current_psf_diameter, current_metric, current_p_value, simulated_data = run_psf_simulation(
        tel_model,
        site_model,
        args_dict,
        current_params,
        data_to_plot,
        radius,
        pdf_pages=pdf_pages if args_dict.get("plot_all", False) else None,
        is_best=False,
        use_ks_statistic=False,
    )

    results.append(
        (
            current_params.copy(),
            current_metric,
            current_p_value,
            current_psf_diameter,
            simulated_data,
        )
    )
    best_metric, best_params, best_psf_diameter = (
        current_metric,
        current_params.copy(),
        current_psf_diameter,
    )

    logger.info(f"Initial RMSD: {current_metric:.6f}, PSF diameter: {current_psf_diameter:.6f} cm")

    iteration = 0
    max_total_iterations = 100

    while iteration < max_total_iterations:
        if current_metric <= rmsd_threshold:
            logger.info(
                f"Optimization converged: RMSD {current_metric:.6f} <= "
                f"threshold {rmsd_threshold:.6f}"
            )
            break

        iteration += 1
        logger.info(f"Gradient descent iteration {iteration}")

        step_result = _perform_gradient_step_with_retries(
            tel_model,
            site_model,
            args_dict,
            current_params,
            current_metric,
            data_to_plot,
            radius,
            learning_rate,
        )
        (
            new_params,
            new_psf_diameter,
            new_metric,
            new_p_value,
            new_simulated_data,
            step_accepted,
            learning_rate,
        ) = step_result

        if not step_accepted or new_params is None:
            learning_rate *= 2.0
            logger.info(f"No step accepted, increasing learning rate to {learning_rate:.6f}")
            continue

        # Step was accepted - update state
        current_params, current_metric, current_psf_diameter = (
            new_params,
            new_metric,
            new_psf_diameter,
        )
        results.append(
            (current_params.copy(), current_metric, None, current_psf_diameter, new_simulated_data)
        )

        if current_metric < best_metric:
            best_metric, best_params, best_psf_diameter = (
                current_metric,
                current_params.copy(),
                current_psf_diameter,
            )

        _create_step_plot(
            pdf_pages,
            args_dict,
            data_to_plot,
            current_params,
            new_psf_diameter,
            new_metric,
            new_p_value,
            new_simulated_data,
        )
        logger.info(f"  Accepted step: improved to {new_metric:.6f}")

    _create_final_plot(
        pdf_pages,
        tel_model,
        site_model,
        args_dict,
        best_params,
        data_to_plot,
        radius,
        best_psf_diameter,
    )
    return best_params, best_psf_diameter, results


def _write_log_interpretation(f, use_ks_statistic):
    """Write interpretation section for the log file."""
    if use_ks_statistic:
        f.write(
            "P-VALUE INTERPRETATION:\n  p > 0.05: Distributions are statistically similar "
            "(good fit)\n"
            "  p < 0.05: Distributions are significantly different (poor fit)\n"
            "  p < 0.01: Very significant difference (very poor fit)\n\n"
        )
    else:
        f.write(
            "RMSD INTERPRETATION:\n  Lower RMSD values indicate better agreement between "
            "measured and simulated PSF curves\n\n"
        )


def _write_iteration_entry(
    f,
    iteration,
    pars,
    metric,
    p_value,
    psf_diameter,
    use_ks_statistic,
    metric_name,
    total_iterations,
    fraction=DEFAULT_FRACTION,
):
    """Write a single iteration entry."""
    status = "FINAL" if iteration == total_iterations - 1 else f"ITER-{iteration:02d}"

    if use_ks_statistic and p_value is not None:
        significance = plot_psf.get_significance_label(p_value)
        label = get_psf_diameter_label(fraction)
        f.write(
            f"[{status}] Iteration {iteration}: KS_stat={metric:.6f}, "
            f"p_value={p_value:.6f} ({significance}), {label}={psf_diameter:.6f} cm\n"
        )
    else:
        label = get_psf_diameter_label(fraction)
        f.write(
            f"[{status}] Iteration {iteration}: {metric_name}={metric:.6f}, "
            f"{label}={psf_diameter:.6f} cm\n"
        )

    for par, value in pars.items():
        f.write(f"    {par}: {_create_log_header_and_format_value(None, None, None, value)}\n")
    f.write("\n")


def _write_optimization_summary(
    f, gd_results, best_pars, best_psf_diameter, metric_name, fraction=DEFAULT_FRACTION
):
    """Write optimization summary section."""
    f.write("OPTIMIZATION SUMMARY:\n")
    best_metric_from_results = min(metric for _, metric, _, _, _ in gd_results)
    f.write(f"Best {metric_name.lower()}: {best_metric_from_results:.6f}\n")

    label = get_psf_diameter_label(fraction)
    f.write(
        f"Best {label}: {best_psf_diameter:.6f} cm\n"
        if best_psf_diameter is not None
        else f"Best {label}: N/A\n"
    )
    f.write(f"Total iterations: {len(gd_results)}\n\nFINAL OPTIMIZED PARAMETERS:\n")
    for par, value in best_pars.items():
        f.write(f"{par}: {_create_log_header_and_format_value(None, None, None, value)}\n")


def write_gradient_descent_log(
    gd_results,
    best_pars,
    best_psf_diameter,
    output_dir,
    tel_model,
    use_ks_statistic=False,
    fraction=DEFAULT_FRACTION,
):
    """
    Write gradient descent optimization progression to a log file.

    Parameters
    ----------
    gd_results : list
        List of tuples containing (params, metric, p_value, psf_diameter, simulated_data)
        for each optimization iteration.
    best_pars : dict
        Dictionary containing the best parameter values found.
    best_psf_diameter : float
        PSF containment diameter in cm for the best parameter set.
    output_dir : Path
        Directory where the log file will be written.
    tel_model : TelescopeModel
        Telescope model object for naming the output file.
    use_ks_statistic : bool, optional
        If True, log KS statistic values; if False, log RMSD values (default: False).
    fraction : float, optional
        PSF containment fraction for labeling (default: 0.8).

    Returns
    -------
    Path
        Path to the created log file.
    """
    metric_name = "KS Statistic" if use_ks_statistic else "RMSD"
    file_suffix = "ks" if use_ks_statistic else "rmsd"
    param_file = output_dir.joinpath(f"psf_gradient_descent_{file_suffix}_{tel_model.name}.log")

    with open(param_file, "w", encoding="utf-8") as f:
        header = _create_log_header_and_format_value(
            f"PSF Parameter Optimization - Gradient Descent Progression ({metric_name})",
            tel_model,
            {"Total iterations": len(gd_results)},
        )
        f.write(header)

        f.write(
            "GRADIENT DESCENT PROGRESSION:\n(Each entry shows the parameters chosen "
            "at each iteration)\n\n"
        )
        _write_log_interpretation(f, use_ks_statistic)

        for iteration, (pars, metric, p_value, psf_diameter, _) in enumerate(gd_results):
            _write_iteration_entry(
                f,
                iteration,
                pars,
                metric,
                p_value,
                psf_diameter,
                use_ks_statistic,
                metric_name,
                len(gd_results),
                fraction,
            )

        _write_optimization_summary(
            f, gd_results, best_pars, best_psf_diameter, metric_name, fraction
        )

    return param_file


def analyze_monte_carlo_error(
    tel_model, site_model, args_dict, data_to_plot, radius, n_simulations=500
):
    """
    Analyze Monte Carlo uncertainty in PSF optimization metrics.

    Runs multiple simulations with the same parameters to quantify the
    statistical uncertainty in the optimization metric due to Monte Carlo
    noise in the ray tracing simulations. Returns None values if no
    measurement data is provided or all simulations fail.

    Parameters
    ----------
    tel_model : TelescopeModel
        Telescope model object with current parameter configuration.
    site_model : SiteModel
        Site model object with environmental conditions.
    args_dict : dict
        Dictionary containing simulation configuration arguments.
    data_to_plot : dict
        Dictionary containing measured PSF data under "measured" key.
    radius : array-like
        Radius values in cm for PSF evaluation.
    n_simulations : int, optional
        Number of Monte Carlo simulations to run (default: 500).

    Returns
    -------
    tuple of (float, float, list, float, float, list, float, float, list)
        - mean_metric: Mean RMSD or KS statistic value
        - std_metric: Standard deviation of metric values
        - metric_values: List of all metric values from simulations
        - mean_p_value: Mean p-value (None if using RMSD)
        - std_p_value: Standard deviation of p-values (None if using RMSD)
        - p_values: List of all p-values from simulations
        - mean_psf_diameter: Mean PSF containment diameter in cm
        - std_psf_diameter: Standard deviation of PSF diameter values
        - psf_diameter_values: List of all PSF diameter values from simulations
    """
    if data_to_plot is None or radius is None:
        logger.error("No PSF measurement data provided. Cannot analyze Monte Carlo error.")
        return None, None, []

    initial_params = get_previous_values(tel_model)
    for param_name, param_values in initial_params.items():
        logger.info(f"  {param_name}: {param_values}")

    use_ks_statistic = args_dict.get("ks_statistic", False)
    metric_values, p_values, psf_diameter_values = [], [], []

    for i in range(n_simulations):
        try:
            psf_diameter, metric, p_value, _ = run_psf_simulation(
                tel_model,
                site_model,
                args_dict,
                initial_params,
                data_to_plot,
                radius,
                use_ks_statistic=use_ks_statistic,
            )
            metric_values.append(metric)
            psf_diameter_values.append(psf_diameter)
            p_values.append(p_value)
        except (ValueError, RuntimeError) as e:
            logger.warning(f"WARNING: Simulation {i + 1} failed: {e}")

    if not metric_values:
        logger.error("All Monte Carlo simulations failed.")
        return None, None, [], None, None, []

    mean_metric, std_metric = np.mean(metric_values), np.std(metric_values, ddof=1)
    mean_psf_diameter, std_psf_diameter = (
        np.mean(psf_diameter_values),
        np.std(psf_diameter_values, ddof=1),
    )

    if use_ks_statistic:
        valid_p_values = [p for p in p_values if p is not None]
        mean_p_value = np.mean(valid_p_values) if valid_p_values else None
        std_p_value = np.std(valid_p_values, ddof=1) if valid_p_values else None
    else:
        mean_p_value = std_p_value = None

    return (
        mean_metric,
        std_metric,
        metric_values,
        mean_p_value,
        std_p_value,
        p_values,
        mean_psf_diameter,
        std_psf_diameter,
        psf_diameter_values,
    )


def write_monte_carlo_analysis(
    mc_results, output_dir, tel_model, use_ks_statistic=False, fraction=DEFAULT_FRACTION
):
    """
    Write Monte Carlo uncertainty analysis results to a log file.

    Parameters
    ----------
    mc_results : tuple
        Tuple of Monte Carlo analysis results from analyze_monte_carlo_error().
    output_dir : Path
        Directory where the log file will be written.
    tel_model : TelescopeModel
        Telescope model object for naming the output file.
    use_ks_statistic : bool, optional
        If True, analyze KS statistic results; if False, analyze RMSD results (default: False).
    fraction : float, optional
        PSF containment fraction for labeling (default: 0.8).

    Returns
    -------
    Path
        Path to the created log file.
    """
    (
        mean_metric,
        std_metric,
        metric_values,
        mean_p_value,
        std_p_value,
        p_values,
        mean_psf_diameter,
        std_psf_diameter,
        psf_diameter_values,
    ) = mc_results

    metric_name = "KS Statistic" if use_ks_statistic else "RMSD"
    file_suffix = "ks" if use_ks_statistic else "rmsd"
    mc_file = output_dir.joinpath(f"monte_carlo_{file_suffix}_analysis_{tel_model.name}.log")

    psf_label = get_psf_diameter_label(fraction)

    with open(mc_file, "w", encoding="utf-8") as f:
        header = _create_log_header_and_format_value(
            f"Monte Carlo {metric_name} Error Analysis",
            tel_model,
            {"Number of simulations": len(metric_values)},
        )
        f.write(header)

        f.write(
            f"MONTE CARLO SIMULATION RESULTS:\nNumber of successful simulations: "
            f"{len(metric_values)}\n\n"
        )
        f.write(f"{metric_name.upper()} STATISTICS:\n")
        f.write(
            f"Mean {metric_name.lower()}: {mean_metric:.6f}\n"
            f"Standard deviation: {std_metric:.6f}\n"
            f"Minimum {metric_name.lower()}: {min(metric_values):.6f}\n"
            f"Maximum {metric_name.lower()}: {max(metric_values):.6f}\n"
            f"Relative error: {(std_metric / mean_metric) * 100:.2f}%\n\n"
        )

        if use_ks_statistic and mean_p_value is not None:
            valid_p_values = [p for p in p_values if p is not None]
            f.write(
                f"P-VALUE STATISTICS:\nMean p-value: {mean_p_value:.6f}\n"
                f"Standard deviation: {std_p_value:.6f}\n"
                f"Minimum p-value: {min(valid_p_values):.6f}\n"
                f"Maximum p-value: {max(valid_p_values):.6f}\n"
                f"Relative error: {(std_p_value / mean_p_value) * 100:.2f}%\n"
            )

            good_fits = sum(1 for p in valid_p_values if p > 0.05)
            fair_fits = sum(1 for p in valid_p_values if 0.01 < p <= 0.05)
            poor_fits = sum(1 for p in valid_p_values if p <= 0.01)
            f.write(
                f"Good fits (p > 0.05): {good_fits}/{len(valid_p_values)} "
                f"({100 * good_fits / len(valid_p_values):.1f}%)\n"
                f"Fair fits (0.01 < p <= 0.05): {fair_fits}/{len(valid_p_values)} "
                f"({100 * fair_fits / len(valid_p_values):.1f}%)\n"
                f"Poor fits (p <= 0.01): {poor_fits}/{len(valid_p_values)} "
                f"({100 * poor_fits / len(valid_p_values):.1f}%)\n\n"
            )

        f.write(
            f"{psf_label} STATISTICS:\nMean {psf_label}: {mean_psf_diameter:.6f} cm\n"
            f"Standard deviation: {std_psf_diameter:.6f} cm\n"
            f"Minimum {psf_label}: {min(psf_diameter_values):.6f} cm\n"
            f"Maximum {psf_label}: {max(psf_diameter_values):.6f} cm\n"
            f"Relative error: {(std_psf_diameter / mean_psf_diameter) * 100:.2f}%\n\n"
        )

        f.write("INDIVIDUAL SIMULATION RESULTS:\n")
        for i, (metric_val, p_value, psf_diameter) in enumerate(
            zip(metric_values, p_values, psf_diameter_values)
        ):
            if use_ks_statistic and p_value is not None:
                if p_value > 0.05:
                    significance = "GOOD"
                elif p_value > 0.01:
                    significance = "FAIR"
                else:
                    significance = "POOR"
                f.write(
                    f"Simulation {i + 1:2d}: {metric_name}={metric_val:.6f}, "
                    f"p_value={p_value:.6f} ({significance}), {psf_label}={psf_diameter:.6f} cm\n"
                )
            else:
                f.write(
                    f"Simulation {i + 1:2d}: {metric_name}={metric_val:.6f}, "
                    f"{psf_label}={psf_diameter:.6f} cm\n"
                )

    return mc_file


def _handle_monte_carlo_analysis(
    tel_model, site_model, args_dict, data_to_plot, radius, output_dir, use_ks_statistic
):
    """Handle Monte Carlo analysis if requested."""
    if not args_dict.get("monte_carlo_analysis", False):
        return False

    mc_results = analyze_monte_carlo_error(tel_model, site_model, args_dict, data_to_plot, radius)
    if mc_results[0] is not None:
        mc_file = write_monte_carlo_analysis(
            mc_results,
            output_dir,
            tel_model,
            use_ks_statistic,
            args_dict.get("fraction", DEFAULT_FRACTION),
        )
        logger.info(f"Monte Carlo analysis results written to {mc_file}")
        mc_plot_file = output_dir.joinpath(f"monte_carlo_uncertainty_{tel_model.name}.pdf")
        plot_psf.create_monte_carlo_uncertainty_plot(
            mc_results, mc_plot_file, args_dict.get("fraction", DEFAULT_FRACTION), use_ks_statistic
        )
    return True


def run_psf_optimization_workflow(tel_model, site_model, args_dict, output_dir):
    """Run the complete PSF parameter optimization workflow using gradient descent."""
    data_to_plot, radius = load_and_process_data(args_dict)
    use_ks_statistic = args_dict.get("ks_statistic", False)

    if _handle_monte_carlo_analysis(
        tel_model, site_model, args_dict, data_to_plot, radius, output_dir, use_ks_statistic
    ):
        return

    # Run gradient descent optimization
    threshold = args_dict.get("rmsd_threshold")
    learning_rate = args_dict.get("learning_rate")

    best_pars, best_psf_diameter, gd_results = run_gradient_descent_optimization(
        tel_model,
        site_model,
        args_dict,
        data_to_plot,
        radius,
        rmsd_threshold=threshold,
        learning_rate=learning_rate,
        output_dir=output_dir,
    )

    # Check if optimization was successful
    if not gd_results or best_pars is None:
        logger.error("Gradient descent optimization failed. No valid results found.")
        if radius is None:
            logger.error(
                "Possible cause: No PSF measurement data provided. "
                "Use --data argument to provide PSF data."
            )
        return

    plot_psf.create_optimization_plots(args_dict, gd_results, tel_model, data_to_plot, output_dir)

    convergence_plot_file = output_dir.joinpath(
        f"gradient_descent_convergence_{tel_model.name}.png"
    )
    plot_psf.create_gradient_descent_convergence_plot(
        gd_results,
        threshold,
        convergence_plot_file,
        args_dict.get("fraction", DEFAULT_FRACTION),
        use_ks_statistic,
    )

    param_file = write_gradient_descent_log(
        gd_results,
        best_pars,
        best_psf_diameter,
        output_dir,
        tel_model,
        use_ks_statistic,
        args_dict.get("fraction", DEFAULT_FRACTION),
    )
    logger.info(f"\nGradient descent progression written to {param_file}")

    plot_psf.create_psf_vs_offaxis_plot(tel_model, site_model, args_dict, best_pars, output_dir)

    if args_dict.get("write_psf_parameters", False):
        logger.info("Exporting best parameters as model files...")
        export_psf_parameters(
            best_pars, args_dict.get("telescope"), args_dict.get("parameter_version"), output_dir
        )



#!/usr/bin/python3

"""
Get list of array layouts or list of elements for a given layout as defined in the db.

To get the list of pre-defined array layouts, use ``--list_available_layouts``.

To get the list of array elements for a given layout, use ``--array_layout_name``.

To get the positions for a set of array elements, use ``--array_element_list``.
Listing of array elements follows this logic:

* explicit listing: e.g., ``-array_element_list MSTN-01, MSTN05``
* listing of types: e.g, ``-array_element_list MSTN`` plots all telescopes of type MSTN.

Command line arguments
----------------------
list_available_layouts : bool, optional
    List available layouts in the database.
array_layout_name : str
    Name of the layout array (e.g., test_layout, alpha, 4mst, etc.).
array_element_list : list
    List of array elements (e.g., telescopes) to plot (e.g., ``LSTN-01 LSTN-02 MSTN``).
coordinate_system : str, optional
    Coordinate system for the array layout (ground or utm).
output_file : str, optional
    Name of the output file to be saved as astropy table (ecsv file)

Examples
--------
List pre-defined array layouts.

.. code-block:: console

    simtools-db-get-array-layouts-from-db --site North --model_version "6.0.0"

Retrieve telescope positions for array layout 'test_layout' from database.

.. code-block:: console

    simtools-db-get-array-layouts-from-db --site North --model_version "6.0.0"
        --array_layout_name test_layout

Retrieve telescope positions from database (utm coordinate system) and write to an ecsv files

.. code-block:: console

    simtools-db-get-array-layouts-from-db --site North --model_version "6.0.0"
      --array_element_list LSTN-01 LSTN-02 MSTN
      --coordinate_system utm
      --output_file telescope_positions-test_layout.ecsv
"""

import simtools.data_model.model_data_writer as writer
from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.model.array_model import ArrayModel
from simtools.model.site_model import SiteModel


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Get list of array elements as defined in the db (array layout).",
    )

    input_group = config.parser.add_mutually_exclusive_group()
    input_group.add_argument(
        "--list_available_layouts",
        help="List available layouts in the database.",
        action="store_true",
        required=False,
    )
    config.parser.add_argument(
        "--coordinate_system",
        help="Coordinate system for the array layout.",
        type=str,
        required=False,
        default="ground",
        choices=["ground", "utm"],
    )
    return config.initialize(
        db_config=True, simulation_model=["site", "layout", "model_version"], output=True
    )


def _layout_from_db(args_dict, db_config):
    """
    Read array elements and their positions from data base using the layout name.

    Parameters
    ----------
    args_dict : dict
        Dictionary with the command line arguments.
    db_config : dict
        Database configuration.

    Returns
    -------
    astropy.table.Table
        Table with array element positions.
    """
    array_model = ArrayModel(
        db_config=db_config,
        model_version=args_dict["model_version"],
        site=args_dict["site"],
        layout_name=args_dict.get("array_layout_name", None),
        array_elements=args_dict.get("array_element_list", None),
    )
    return array_model.export_array_elements_as_table(
        coordinate_system=args_dict["coordinate_system"]
    )


def main():
    """Get list of array layouts or list of elements for a given layout as defined in the db."""
    app_context = startup_application(_parse)

    if app_context.args.get("list_available_layouts", False):
        if app_context.args.get("site", None) is None:
            raise ValueError("Site must be provided to list available layouts.")
        site_model = SiteModel(
            db_config=app_context.db_config,
            model_version=app_context.args["model_version"],
            site=app_context.args["site"],
        )
        print(site_model.get_list_of_array_layouts())
    else:
        app_context.logger.info("Array layout: %s", app_context.args["array_layout_name"])
        layout = _layout_from_db(app_context.args, app_context.db_config)
        layout.pprint()

        if not app_context.args.get("output_file_from_default", False):
            writer.ModelDataWriter.dump(
                args_dict=app_context.args,
                output_file=app_context.args["output_file"],
                metadata=None,
                product_data=layout,
            )


if __name__ == "__main__":
    main()



"""Definition and modeling of a calibration device."""

import logging

from simtools.model.model_parameter import ModelParameter


class CalibrationModel(ModelParameter):
    """
    CalibrationModel represents the MC model of an individual calibration device.

    It provides functionality to read the required parameters from the DB.

    Parameters
    ----------
    site: str
        Site name (e.g., South or North).
    calibration_device_model_name: str
        Calibration device model name (ex. ILLS-01, ILLN-01, ...).
    db_config: dict
        Database configuration.
    model_version: str
        Model version.
    label: str, optional
        Instance label. Important for output file naming.
    overwrite_model_parameters: str, optional
        File name to overwrite model parameters from DB with provided values.
    """

    def __init__(
        self,
        site,
        calibration_device_model_name,
        db_config,
        model_version,
        label=None,
        overwrite_model_parameters=None,
    ):
        """Initialize CalibrationModel."""
        super().__init__(
            site=site,
            array_element_name=calibration_device_model_name,
            collection="calibration_devices",
            db_config=db_config,
            model_version=model_version,
            label=label,
            overwrite_model_parameters=overwrite_model_parameters,
        )

        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init CalibrationModel %s %s", site, calibration_device_model_name)



#!/usr/bin/python3
"""Helper functions related to model parameters."""

import math

from simtools.model.calibration_model import CalibrationModel
from simtools.model.site_model import SiteModel
from simtools.model.telescope_model import TelescopeModel
from simtools.utils import names


def initialize_simulation_models(
    label, db_config, model_version, site, telescope_name, calibration_device_name=None
):
    """
    Initialize simulation models for a single telescope, site, and calibration device model.

    Parameters
    ----------
    label: str
        Label for the simulation.
    db_config: dict
        Database configuration.
    model_version: str
        Version of the simulation model
    site: str
        Name of the site.
    telescope_name: str
        Name of the telescope.
    calibration_device_name: str, optional
        Name of the calibration device.

    Returns
    -------
    Tuple
        Tuple containing the telescope site, (optional) calibration device model.
    """
    tel_model = TelescopeModel(
        site=site,
        telescope_name=telescope_name,
        db_config=db_config,
        model_version=model_version,
        label=label,
    )
    site_model = SiteModel(
        site=site,
        model_version=model_version,
        db_config=db_config,
        label=label,
    )
    if calibration_device_name is not None:
        calibration_model = CalibrationModel(
            site=site,
            calibration_device_model_name=calibration_device_name,
            db_config=db_config,
            model_version=model_version,
            label=label,
        )
    else:
        calibration_model = None
    for model in tel_model, site_model:
        model.export_model_files()
    return tel_model, site_model, calibration_model


def compute_telescope_transmission(pars: list[float], off_axis: float) -> float:
    """
    Compute telescope transmission (0 < T < 1) for a given off-axis angle.

    The telescope transmission depends on the MC model used.

    Parameters
    ----------
    pars: list of float
        Parameters of the telescope transmission. Len(pars) should be 5 or 6.
    off_axis: float
        Off-axis angle in deg.

    Returns
    -------
    float
        Telescope transmission.
    """
    _deg_to_rad = math.pi / 180.0
    if pars[1] == 0:
        return pars[0]

    t = math.sin(off_axis * _deg_to_rad) / (pars[3] * _deg_to_rad)
    return pars[0] / (1.0 + pars[2] * t ** pars[4])


def is_two_mirror_telescope(telescope_model_name: str) -> bool:
    """
    Determine if the telescope model is a two-mirror telescope.

    Parameters
    ----------
    telescope_model_name: str
        Name of the telescope model.

    Returns
    -------
    bool
        True if it is a two-mirror telescope.
    """
    tel_type = names.get_array_element_type_from_name(telescope_model_name)
    if "SST" in tel_type or "SCT" in tel_type:
        return True
    return False



"""Value and quantity conversion."""

import logging
import re

import numpy as np
from astropy import units as u

import simtools.utils.general as gen

_logger = logging.getLogger(__name__)


def extract_type_of_value(value) -> str:
    """
    Extract the string representation of the the type of a value.

    For example, for a string, it returns 'str' rather than '<class 'str'>'.
    Take into account also the case where the value is a numpy type.
    """
    _type = str(type(value))
    if "numpy" in _type:
        return re.sub(r"\d+", "", _type.split("'")[1].split(".")[-1])
    if "astropy" in _type:
        raise NotImplementedError("Astropy types are not supported yet.")
    return _type.split("'")[1]


def get_value_unit_type(value, unit_str=None):
    """
    Get the value, unit and type of a value.

    The value is stripped of its unit and the unit is returned
    in its string form (i.e., to_string()).
    The type is returned as a string representation of the type.
    For example, for a string, it returns 'str' rather than '<class 'str'>'.
    An additional unit string can be given and the return value is converted to this units.

    Note that Quantities are always floats, even if the original value is represented as an int.

    Parameters
    ----------
    value: str, int, float, bool, u.Quantity
        Value to be parsed.
    unit_str: str
        Unit to be used for the value.

    Returns
    -------
    type of value, str, str
        Value, unit in string representation (to_string())),
        and string representation of the type of the value.
    """
    base_unit = None
    if isinstance(value, str | u.Quantity):
        try:
            _quantity_value = u.Quantity(value)
            base_value = _quantity_value.value
            base_type = extract_type_of_value(base_value)
            if _quantity_value.unit.to_string() != "":
                base_unit = _quantity_value.unit.to_string()
                try:  # handle case of e.g., "0 0" and avoid unit.scale
                    float(base_unit)
                    base_value = value
                    base_type = "str"
                    base_unit = None
                except ValueError:
                    pass
        # ValueError: covers strings of type "5 not a unit"
        except (TypeError, ValueError):
            base_value = value
            base_type = "str"
    else:
        base_value = value
        base_type = extract_type_of_value(base_value)

    if unit_str is not None:
        try:
            base_value = base_value * u.Unit(base_unit).to(u.Unit(unit_str))
        except u.UnitConversionError:
            _logger.error(f"Cannot convert {base_unit} to {unit_str}.")
            raise
        except TypeError:
            pass
        base_unit = unit_str

    return base_value, _unit_as_string(base_unit), base_type


def split_value_and_unit(value, is_integer=False):
    """
    Split a value into its value and unit.

    Takes into account the case where the value is a Quantity, a number,
    or a simtools-type string encoding a list of values and units.

    Parameters
    ----------
    value: str, int, float, bool, u.Quantity
        Value to be parsed.
    is_integer: bool
        Flag to indicate if the value is an integer.

    Returns
    -------
    value, str
        Value and units as (value, unit), or lists of values and unites
    """
    if isinstance(value, u.Quantity):
        return _split_value_is_quantity(value, is_integer)
    if isinstance(value, str):
        return _split_value_is_string(value, is_integer)
    if isinstance(value, list | np.ndarray):
        return _split_value_is_list(value, is_integer)
    return value, None


def _split_value_is_quantity(value, is_integer=False):
    """Split value and unit for an astropy Quantity."""
    if isinstance(value.value, list | np.ndarray):  # type [100.0, 200] * u.m
        return [int(v) if is_integer else v for v in value.value], [str(value.unit)] * len(value)
    return int(value.value) if is_integer else value.value, str(value.unit)


def _split_value_is_string(value, is_integer=False):
    """Split value and unit for a string."""
    if value.isdigit():  # single integer value
        return int(value), None
    try:  # single value with/without unit
        quantity = u.Quantity(value)
        unit = str(quantity.unit)
        if unit.isdigit():  # cases where numbers are wrongly identified as units
            raise ValueError
        return (int(quantity.value), unit) if is_integer else (quantity.value, unit)
    except ValueError:
        return _split_value_is_list(gen.convert_string_to_list(value), is_integer)
    except TypeError:  # string value (not numerical)
        return value, None


def _split_value_is_list(value, is_integer=False):
    """Split value and unit for a list."""
    value_list = []
    unit_list = []
    for item in value:
        _value, _unit = split_value_and_unit(item)
        value_list.append(int(_value) if is_integer else _value)
        if isinstance(_unit, str):
            unit_list.append(_unit)
        else:
            unit_list.append(None)
    return value_list, unit_list


def get_value_as_quantity(value, unit):
    """
    Get a value as a Quantity with a given unit. If value is a Quantity, convert to the given unit.

    Parameters
    ----------
    value:
        value to get a unit. It can be a float, int, or a Quantity (convertible to 'unit').
    unit: astropy.units.Unit
        Unit to apply to 'quantity'.

    Returns
    -------
    astropy.units.Quantity
        Quantity of value 'quantity' and unit 'unit'.

    Raises
    ------
    ValueError
        If the value cannot be converted to the given unit.
    """
    if isinstance(value, u.Quantity):
        try:
            return value.to(unit)
        except u.UnitConversionError as exc:
            raise ValueError(f"Cannot convert {value} with unit {value.unit} to {unit}.") from exc
    elif not isinstance(value, int | float):
        return value

    if unit is None or unit == "null":
        return value * u.dimensionless_unscaled

    return value * u.Unit(unit)


def _unit_as_string(unit):
    """Return the string representation of a unit. Collapse if it is a list of identical items."""
    if not isinstance(unit, list):
        unit = [unit]
    unit = [str(element) if element is not None else None for element in unit]
    return unit[0] if len(set(unit)) == 1 else unit



#!/usr/bin/python3

r"""
Merge CORSIKA limit tables from multiple grid points and check grid completeness.

The corsika limit tables are first generated by the application
'simtools-production-derive-corsika-limits' and contain derived limits for
energy, radial distance, and viewcone for different observational conditions.
The individual tables are merged into a single table for each grid point by
the application 'merge_tables'. This tool finalizes the merging process and
checks the completeness of the grid against a provided grid definition file.

This tool supports three main use cases:

1. Merge multiple CORSIKA limit tables into a single file and optionally generate
   plots of the derived limits.
2. Merge tables and also check for grid completeness against a provided grid
   definition file. This requires the --grid_definition parameter. Coverage plots
   can also be generated.
3. Check grid completeness of an already merged table file. This requires both
   the --merged_table and --grid_definition parameters.

Command line arguments
----------------------
input_files (str)
    Directory containing corsika_simulation_limits_lookup*.ecsv files to be merged,
    or a list of specific files to merge. Not used if --merged_table is provided.
input_files_list (str)
    Path to a text file containing a list of input files (one file path per line)
    to be merged. Not used if --merged_table is provided.
merged_table (str)
    Path to an already merged table file. Used for checking grid completeness.
grid_definition (str)
    Path to a YAML file defining the expected grid points. Required for grid
    completeness checks and coverage plots.
output_file (str, optional)
    Name of the output file for the merged limits table.
    Default is "merged_corsika_limits.ecsv".
plot_grid_coverage (bool, optional)
    Flag to generate plots showing grid coverage. Requires --grid_definition.
plot_limits (bool, optional)
    Flag to generate plots showing the derived limits.

Examples
--------
1. Merge CORSIKA limit tables from a directory:

   .. code-block:: console

      simtools-production-merge-corsika-limits \\
          --input_files "simtools-output/corsika_limits/" \\
          --output_file merged_limits.ecsv --plot_limits

2. Merge tables and check grid completeness:

   .. code-block:: console

      simtools-production-merge-corsika-limits \\
          --input_files "simtools-output/corsika_limits/" \\
          --grid_definition grid_definition.yaml \\
          --output_file merged_limits.ecsv --plot_grid_coverage

3. Check grid completeness of an existing merged table:

   .. code-block:: console

      simtools-production-merge-corsika-limits \\
          --merged_table merged_limits.ecsv \\
          --grid_definition grid_definition.yaml --plot_grid_coverage

4. Merge tables using a list of files from a text file:

   .. code-block:: console

      simtools-production-merge-corsika-limits \\
          --input_files_list file_list.txt \\
          --output_file merged_limits.ecsv
"""

from pathlib import Path

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.data_model import data_reader
from simtools.io import ascii_handler
from simtools.production_configuration.merge_corsika_limits import CorsikaMergeLimits


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Merge CORSIKA limit tables and check grid completeness.",
    )
    config.parser.add_argument(
        "--input_files",
        type=str,
        default=None,
        nargs="+",
        help=(
            "A list of input files to be merged, or a single directory "
            "containing the files (*.ecsv)."
        ),
    )
    config.parser.add_argument(
        "--input_files_list",
        type=str,
        default=None,
        help=(
            "Path to a text file containing a list of input files (one file path per line) "
            "to be merged."
        ),
    )
    config.parser.add_argument(
        "--merged_table",
        type=str,
        default=None,
        help="Path to an already merged table file.",
    )
    config.parser.add_argument(
        "--grid_definition",
        type=str,
        default=None,
        help="Path to YAML file defining the expected grid points.",
    )
    config.parser.add_argument(
        "--plot_grid_coverage",
        help="Generate plots showing grid coverage.",
        action="store_true",
        default=False,
    )
    config.parser.add_argument(
        "--plot_limits",
        help="Generate plots showing the derived limits.",
        action="store_true",
        default=False,
    )
    return config.initialize(output=True)


def main():
    """Merge CORSIKA limit tables and check grid completeness."""
    app_context = startup_application(_parse)

    merger = CorsikaMergeLimits()
    grid_definition = (
        ascii_handler.collect_data_from_file(app_context.args["grid_definition"])
        if app_context.args.get("grid_definition")
        else None
    )

    if app_context.args.get("merged_table"):
        # Case 3: Check coverage on an existing merged table
        merged_table_path = Path(app_context.args["merged_table"]).expanduser()
        merged_table = data_reader.read_table_from_file(merged_table_path)
        input_files = [merged_table_path]
    elif app_context.args.get("input_files") or app_context.args.get("input_files_list"):
        # Case 1 & 2: Merge files
        input_files = []

        # Process input_files argument
        if app_context.args.get("input_files"):
            raw_paths = app_context.args.get("input_files")
            if len(raw_paths) == 1 and Path(raw_paths[0]).expanduser().is_dir():
                input_dir = Path(raw_paths[0]).expanduser()
                input_files.extend(input_dir.glob("*.ecsv"))
            else:
                input_files.extend(Path(f).expanduser() for f in raw_paths)

        # Process input_files_list argument
        if app_context.args.get("input_files_list"):
            files_from_list = merger.read_file_list(app_context.args["input_files_list"])
            input_files.extend(files_from_list)

        if not input_files:
            raise FileNotFoundError(
                "No input files found. Check --input_files or --input_files_list arguments."
            )
        merged_table = merger.merge_tables(input_files)
    else:
        raise ValueError(
            "Either --input_files, --input_files_list, or --merged_table must be provided."
        )

    is_complete, grid_completeness = merger.check_grid_completeness(merged_table, grid_definition)

    if app_context.args.get("plot_grid_coverage"):
        merger.plot_grid_coverage(merged_table, grid_definition)

    if app_context.args.get("plot_limits"):
        merger.plot_limits(merged_table)

    if not app_context.args.get("merged_table"):
        # Write output file only when merging
        output_file = merger.output_dir / app_context.args["output_file"]
        merger.write_merged_table(
            merged_table,
            output_file,
            input_files,
            {
                "is_complete": is_complete,
                "expected": grid_completeness.get("expected", 0),
                "found": grid_completeness.get("found", 0),
                "missing": grid_completeness.get("missing", []),
            },
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""Functions for plotting pixel layout information."""

import logging
from pathlib import Path

import astropy.units as u
import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection

from simtools.db import db_handler
from simtools.io import io_handler
from simtools.model.model_utils import is_two_mirror_telescope
from simtools.utils import names
from simtools.visualization import legend_handlers as leg_h
from simtools.visualization import visualize

logger = logging.getLogger(__name__)


def plot(config, output_file, db_config=None):
    """
    Plot pixel layout based on configuration.

    Parameters
    ----------
    config : dict
        Configuration dictionary containing:
        - file_name : str, name of camera config file
        - column_x : str, x-axis label
        - column_y : str, y-axis label
        - parameter_version: str, version of the parameter
        - telescope : str, name of the telescope
    output_file : str
        Path where to save the plot
    db_config : dict, optional
        Database configuration.

    Returns
    -------
    None
        The function saves the plot to the specified output file.
    """
    db = db_handler.DatabaseHandler(db_config=db_config)
    db.export_model_file(
        parameter=config["parameter"],
        site=config["site"],
        array_element_name=config.get("telescope"),
        parameter_version=config.get("parameter_version"),
        model_version=config.get("model_version"),
        export_file_as_table=False,
    )
    data_file_path = Path(io_handler.IOHandler().get_output_directory() / f"{config['file_name']}")
    fig = plot_pixel_layout_from_file(
        data_file_path,
        config["telescope"],
        pixels_id_to_print=80,
    )
    visualize.save_figure(fig, output_file)
    plt.close(fig)


def plot_pixel_layout_from_file(dat_file_path, telescope_model_name, **kwargs):
    """
    Plot the pixel layout from a camera config file.

    This function reads the pixel configuration from the specified camera config file and
    generates a plot of the pixel layout for the given telescope model.

    Parameters
    ----------
    dat_file_path : str or Path
        Path to the camera config file containing pixel configuration
    telescope_model_name : str
        Name/model of the telescope
    **kwargs
        pixels_id_to_print : int
            Number of pixel IDs to print in the plot
        title : str
            Plot title
        xtitle : str
            X-axis label
        ytitle : str
            Y-axis label

    Returns
    -------
    matplotlib.figure.Figure
        The generated figure
    """
    logger.info(f"Plotting pixel layout for {telescope_model_name} from {dat_file_path}")

    pixel_data = _prepare_pixel_data(
        dat_file_path,
        telescope_model_name,
    )

    return _create_pixel_plot(
        pixel_data,
        telescope_model_name,
        pixels_id_to_print=kwargs.get("pixels_id_to_print", 50),
        title=kwargs.get("title"),
        xtitle=kwargs.get("xtitle"),
        ytitle=kwargs.get("ytitle"),
    )


def _prepare_pixel_data(dat_file_path, telescope_model_name):
    """Prepare pixel data from sim_telarray camera configuration file.

    This function reads the pixel configuration from the specified camera config file and
    prepares the data for plotting, including applying any necessary rotations.

    Parameters
    ----------
    dat_file_path : str or Path
        Path to the camera config file containing pixel configuration
    telescope_model_name : str
        Name/model of the telescope

    Returns
    -------
    dict
        Dictionary containing pixel data
    """
    config = _read_pixel_config(dat_file_path)
    x_pos = np.array(config["x"])
    y_pos = np.array(config["y"])

    if not is_two_mirror_telescope(telescope_model_name):
        y_pos = -y_pos

    rotate_angle = (
        config.get("rotate_angle") if config.get("rotate_angle") is not None else (0.0 * u.deg)
    )

    # Apply telescope-specific adjustments
    if "SST" in telescope_model_name or "SCT" in telescope_model_name:
        total_rotation = (90 * u.deg) - (rotate_angle)
    else:
        total_rotation = (-90 * u.deg) - (rotate_angle)

    # Apply rotation
    rot_angle = total_rotation.to(u.rad).value
    x_rot = x_pos * np.cos(rot_angle) - y_pos * np.sin(rot_angle)
    y_rot = y_pos * np.cos(rot_angle) + x_pos * np.sin(rot_angle)
    x_pos, y_pos = x_rot, y_rot

    return {
        "x": x_pos,
        "y": y_pos,
        "pixel_ids": config["pixel_ids"],
        "pixels_on": config["pixels_on"],
        "pixel_shape": config["pixel_shape"],
        "pixel_diameter": config["pixel_diameter"],
        "pixel_spacing": config["pixel_spacing"],
        "module_number": config["module_number"],
        "module_gap": config["module_gap"],
        "rotation": total_rotation,
    }


def _create_pixel_plot(
    pixel_data, telescope_model_name, pixels_id_to_print=50, title=None, xtitle=None, ytitle=None
):
    """
    Create and configure the pixel layout plot.

    Parameters
    ----------
    pixel_data : dict
        Dictionary containing pixel configuration data
    telescope_model_name : str
        Name of telescope model
    pixels_id_to_print : int, optional
        Number of pixel IDs to print, default 50
    title : str, optional
        Plot title
    xtitle : str, optional
        X-axis label
    ytitle : str, optional
        Y-axis label

    Returns
    -------
    matplotlib.figure.Figure
        The generated figure
    """
    fig, ax = plt.subplots(figsize=(8, 8))

    # Create patches
    on_pixels, edge_pixels, off_pixels = _create_pixel_patches(
        pixel_data["x"],
        pixel_data["y"],
        pixel_data["pixel_diameter"],
        pixel_data["module_number"],
        pixel_data["module_gap"],
        pixel_data["pixel_spacing"],
        pixel_data["pixel_shape"],
        pixel_data["pixels_on"],
        pixel_data["pixel_ids"],
        pixels_id_to_print,
        telescope_model_name,
    )

    # Combine all patches into a single collection
    all_patches = on_pixels + edge_pixels + off_pixels
    facecolors = [
        "none"
        if i < len(on_pixels)
        else (*mcolors.to_rgb("brown"), 0.5)
        if i < len(on_pixels) + len(edge_pixels)
        else "black"
        for i in range(len(on_pixels) + len(edge_pixels) + len(off_pixels))
    ]
    edgecolors = (
        ["black"] * len(on_pixels)
        + [(*mcolors.to_rgb("black"), 1)] * len(edge_pixels)
        + ["black"] * len(off_pixels)
    )
    linewidths = [0.2] * len(all_patches)

    # Add the combined collection
    ax.add_collection(
        PatchCollection(
            all_patches,
            facecolor=facecolors,
            edgecolor=edgecolors,
            linewidth=linewidths,
            match_original=True,
        )
    )

    # Configure plot with titles
    _configure_plot(
        ax,
        pixel_data["x"],
        pixel_data["y"],
        rotation=pixel_data["rotation"],
        title=title,
        xtitle=xtitle,
        ytitle=ytitle,
    )
    _add_legend(ax, on_pixels, off_pixels)

    return fig


def _read_pixel_config(dat_file_path):
    """Read pixel configuration from a camera configuration file.

    This function reads the pixel configuration from the specified camera config file and
    returns it as a dictionary. It parses information such as pixel positions,
    module numbers, and other relevant parameters.

    Parameters
    ----------
    dat_file_path : str or Path
        Path to the camera config file containing pixel configuration

    Returns
    -------
    dict
        config containing pixel data
    """
    config = {
        "x": [],
        "y": [],
        "pixel_ids": [],
        "pixels_on": [],
        "pixel_shape": None,
        "pixel_diameter": None,
        "pixel_spacing": None,
        "module_gap": None,
        "trigger_groups": [],
        "rotate_angle": None,
        "module_number": [],
    }

    with open(dat_file_path, encoding="utf-8") as f:
        for line in f:
            line = line.strip()

            if not line:
                continue

            # Parse specific information from the file
            if line.startswith("Rotate"):
                # Parse rotation angle from line like "Rotate 10.893" (u.deg)
                config["rotate_angle"] = float(line.split()[1].strip()) * u.deg

            elif line.startswith("PixType"):
                parts = line.split()
                config["pixel_shape"] = int(parts[5].strip())
                config["pixel_diameter"] = float(parts[6].strip())

            elif "Pixel spacing is" in line:
                config["pixel_spacing"] = float(line.split("spacing is")[1].strip().split()[0])

            elif "Between modules is an additional gap of" in line:
                config["module_gap"] = float(line.split("gap of")[1].strip().split()[0])

            elif line.startswith("Pixel"):
                parts = line.split()
                config["x"].append(float(parts[3].strip()))
                config["y"].append(float(parts[4].strip()))
                config["module_number"].append(float(parts[5].strip()))
                config["pixel_ids"].append(int(parts[1].strip()))
                config["pixels_on"].append(int(parts[9].strip()) != 0)

    config["pixel_spacing"] = (
        config["pixel_diameter"] if config["pixel_spacing"] is None else config["pixel_spacing"]
    )
    config["module_gap"] = 0.0 if config["module_gap"] is None else config["module_gap"]

    return config


def _create_patch(x, y, diameter, shape):
    """Create a single matplotlib patch for a pixel.

    This function creates a matplotlib patch (shape) for a single pixel based on
    its position, diameter, and shape type. Supported shapes are circles, squares,
    and hexagons.

    Parameters
    ----------
    x, y : float
        Center coordinates of the pixel
    diameter : float
        Diameter of the pixel
    shape : int
        Pixel shape type:
        0: circular
        1: hexagonal (flat x)
        2: square
        3: hexagonal (flat y)

    Returns
    -------
    matplotlib.patches.Patch
        The created patch object for the pixel
    """
    if shape == 0:  # Circular
        return mpatches.Circle((x, y), radius=diameter / 2)
    if shape in (1, 3):  # Hexagonal
        return mpatches.RegularPolygon(
            (x, y),
            numVertices=6,
            radius=diameter / np.sqrt(3),
            orientation=np.deg2rad(30 if shape == 3 else 0),
        )
    # Square
    return mpatches.Rectangle((x - diameter / 2, y - diameter / 2), width=diameter, height=diameter)


def _is_edge_pixel(
    x, y, x_pos, y_pos, module_ids, pixel_spacing, module_gap, shape, current_module_id
):
    """
    Determine if a pixel is on the edge based on neighbor count.

    Parameters
    ----------
    x, y : float
        Coordinates of the pixel being checked.
    x_pos, y_pos : array-like
        Arrays of x and y positions of all pixels.
    module_ids : array-like
        Array of module IDs corresponding to each pixel.
    pixel_spacing : float
        Center-to-center spacing between pixels.
    module_gap : float
        Additional gap between modules.
    shape : int
        Pixel shape type (0: circular, 1/3: hexagonal, 2: square).
    current_module_id : int
        Module ID of the current pixel.

    Returns
    -------
    bool
        True if the pixel is an edge pixel, False otherwise.
    """
    # Determine the maximum number of neighbors based on the pixel shape
    if shape == 0:  # Circular
        max_neighbors = 8
    elif shape in (1, 3):  # Hexagonal
        max_neighbors = 6
    elif shape == 2:  # Square
        max_neighbors = 4
    else:
        raise ValueError(f"Unsupported pixel shape: {shape}")

    neighbor_count = _count_neighbors(
        x, y, x_pos, y_pos, module_ids, pixel_spacing, module_gap, current_module_id
    )

    # A pixel is an edge pixel if it has fewer neighbors than the maximum
    return neighbor_count < max_neighbors


def _create_pixel_patches(
    x_pos,
    y_pos,
    diameter,
    module_number,
    module_gap,
    spacing,
    shape,
    pixels_on,
    pixel_ids,
    pixels_id_to_print,
    telescope_model_name,
):
    """Create matplotlib patches for different pixel types.

    This function creates the matplotlib patches (shapes) for all pixels in the
    layout, categorizing them into "on", "edge", and "off" pixels based on their
    status and position.

    Parameters
    ----------
    x_pos, y_pos : array-like
        X and Y coordinates of the pixel centers
    diameter : float
        Diameter of the pixels
    module_number : array-like
        Module numbers for each pixel
    module_gap : float
        Gap between modules
    spacing : float
        Pixel spacing
    shape : array-like
        Shape types for each pixel
    pixels_on : array-like
        Status indicating if each pixel is "on"
    pixel_ids : array-like
        Unique IDs for each pixel
    pixels_id_to_print : int
        Number of pixel IDs to print on the plot
    telescope_model_name : str
        Name of the telescope model

    Returns
    -------
    tuple
        Three lists of patches for "on", "edge", and "off" pixels
    """
    on_pixels, edge_pixels, off_pixels = [], [], []

    array_element_type = names.get_array_element_type_from_name(telescope_model_name)
    font_size = 2 if "SCT" in array_element_type else 4

    for i, (x, y) in enumerate(zip(x_pos, y_pos)):
        patch = _create_patch(x, y, diameter, shape)

        if pixels_on[i]:
            if _is_edge_pixel(
                x, y, x_pos, y_pos, module_number, spacing, module_gap, shape, module_number[i]
            ):
                edge_pixels.append(patch)
            else:
                on_pixels.append(patch)
        else:
            off_pixels.append(patch)

        if pixel_ids[i] < pixels_id_to_print:
            plt.text(x, y, pixel_ids[i], ha="center", va="center", fontsize=font_size)

    return on_pixels, edge_pixels, off_pixels


def _count_neighbors(x, y, x_pos, y_pos, module_ids, pixel_spacing, module_gap, current_module_id):
    """
    Count the number of neighboring pixels within the appropriate distance.

    Parameters
    ----------
    x, y : float
        Coordinates of the pixel being checked.
    x_pos, y_pos : array-like
        Arrays of x and y positions of all pixels.
    module_ids : array-like
        Array of module IDs corresponding to each pixel.
    pixel_spacing : float
        Center-to-center spacing between pixels.
    module_gap : float
        Additional gap between modules.
    current_module_id : int
        Module ID of the current pixel.

    Returns
    -------
    int
        Number of neighboring pixels.
    """
    count = 0
    tolerance = 1e-6

    for x2, y2, module_id2 in zip(x_pos, y_pos, module_ids):
        # Skip the pixel itself
        if x == x2 and y == y2:
            continue

        # Calculate the distance between the current pixel and the potential neighbor
        dist = np.sqrt((x - x2) ** 2 + (y - y2) ** 2)

        # Determine max distance based on whether pixels are in same module
        max_distance = (
            pixel_spacing + (0 if current_module_id == module_id2 else module_gap) + tolerance
        ) * 1.2

        if dist <= max_distance:
            count += 1

    return count


def _configure_plot(
    ax,
    x_pos,
    y_pos,
    rotation=0 * u.deg,
    title=None,
    xtitle=None,
    ytitle=None,
):
    """Configure the plot with titles, labels, and limits.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to configure
    x_pos, y_pos : array-like
        Arrays of x and y positions of pixels
    rotation : Astropy quantity in degrees, optional
        Rotation angle in degrees, default 0
    title : str, optional
        Plot title
    xtitle : str, optional
        X-axis label
    ytitle : str, optional
        Y-axis label


    Returns
    -------
    None
        The function modifies the plot axes in place.
    """
    # First set the aspect ratio
    ax.set_aspect("equal")

    # Calculate the axis limits
    x_min, x_max = min(x_pos), max(x_pos)
    y_min, y_max = min(y_pos), max(y_pos)

    # Add some padding
    x_padding = (x_max - x_min) * 0.1
    y_padding = (y_max - y_min) * 0.1

    # Set limits with padding
    ax.set_xlim(x_min - x_padding, x_max + x_padding)
    ax.set_ylim(y_min - y_padding, y_max + y_padding)

    plt.grid(True)
    ax.set_axisbelow(True)

    plt.xlabel(xtitle or "Horizontal scale [cm]", fontsize=18, labelpad=0)
    plt.ylabel(ytitle or "Vertical scale [cm]", fontsize=18, labelpad=0)
    ax.set_title(
        title or "Pixel layout",
        fontsize=15,
        y=1.02,
    )
    plt.tick_params(axis="both", which="major", labelsize=15)

    _add_coordinate_axes(ax, rotation)
    x_min = min(x_pos) - (max(x_pos) - min(x_pos)) * 0.05
    y_min = min(y_pos) - (max(y_pos) - min(y_pos)) * 0.05
    ax.text(x_min, y_min, "For an observer facing the camera", fontsize=10, ha="left", va="bottom")


def _add_coordinate_axes(ax, rotation=0 * u.deg):
    """Add coordinate system axes to the plot."""
    # Setup dimensions and positions
    x_min, x_max = ax.get_xlim()
    y_min, y_max = ax.get_ylim()
    plot_size = min(x_max - x_min, y_max - y_min)
    axis_length = plot_size * 0.08

    x_origin = x_max - axis_length * 1.0
    y_origin_az = y_min + axis_length * 2.5
    y_origin_pix = y_min + axis_length * 1.2

    arrow_style = {
        "head_width": axis_length * 0.15,
        "head_length": axis_length * 0.15,
        "width": axis_length * 0.02,
    }
    arrow_length = 0.6
    is_sst = abs(rotation - (90.0 * u.deg)).value < 1.0
    az_direction = 1 if is_sst else -1

    def add_arrow_label(ox, oy, dx, dy, label, offset, color="black", ha="center", va="center"):
        """Adding arrows with label."""
        ax.arrow(ox, oy, dx, dy, fc=color, ec=color, **arrow_style)
        if np.sqrt(dx**2 + dy**2) > 0:  # If not zero vector
            dir_unit = np.sqrt(dx**2 + dy**2)
            ax.text(
                ox + dx + dx / dir_unit * axis_length * offset,
                oy + dy + dy / dir_unit * axis_length * offset,
                label,
                ha=ha,
                va=va,
                color=color,
                fontsize=10,
                fontweight="bold",
            )

    # Az-Alt axes
    az_dx = az_direction * axis_length * arrow_length
    add_arrow_label(
        x_origin,
        y_origin_az,
        az_dx,
        0,
        "Az",
        0.25,
        "red",
        ha="left" if az_direction > 0 else "right",
    )
    add_arrow_label(
        x_origin, y_origin_az, 0, -axis_length * arrow_length, "Alt", 0.25, "red", va="top"
    )

    # Pixel coordinate axes
    rot_angle = rotation.to(u.rad).value
    x_direction = -1 if is_sst else 1
    x_dir = x_direction * axis_length * arrow_length * np.cos(rot_angle)
    y_dir = x_direction * axis_length * arrow_length * np.sin(rot_angle)
    add_arrow_label(x_origin, y_origin_pix, x_dir, y_dir, "$\\mathrm{x}_\\mathrm{pix}$", 0.45)

    y_dx = axis_length * arrow_length * np.sin(rot_angle)
    y_dy = -axis_length * arrow_length * np.cos(rot_angle)
    add_arrow_label(x_origin, y_origin_pix, y_dx, y_dy, "$\\mathrm{y}_\\mathrm{pix}$", 0.45)


def _add_legend(ax, on_pixels, off_pixels):
    """Add legend to the plot."""
    legend_objects = [leg_h.PixelObject(), leg_h.EdgePixelObject()]
    legend_labels = ["Pixel", "Edge pixel"]

    # Choose handler based on pixel shape
    is_hex = isinstance(on_pixels[0], mpatches.RegularPolygon)
    legend_handler_map = {
        leg_h.PixelObject: leg_h.HexPixelHandler() if is_hex else leg_h.SquarePixelHandler(),
        leg_h.EdgePixelObject: leg_h.HexEdgePixelHandler()
        if is_hex
        else leg_h.SquareEdgePixelHandler(),
        leg_h.OffPixelObject: leg_h.HexOffPixelHandler()
        if is_hex
        else leg_h.SquareOffPixelHandler(),
    }

    if off_pixels:
        legend_objects.append(leg_h.OffPixelObject())
        legend_labels.append("Disabled pixel")

    ax.legend(
        legend_objects,
        legend_labels,
        handler_map=legend_handler_map,
        prop={"size": 11},
        loc="upper right",
    )



"""Application control utilities for startup and shutdown simtools applications."""

import logging
from dataclasses import dataclass
from pathlib import Path

import simtools.utils.general as gen
from simtools.io import io_handler


@dataclass
class ApplicationContext:
    """Container for common application context elements."""

    args: dict
    db_config: dict
    logger: logging.Logger
    io_handler: io_handler.IOHandler | None


def startup_application(parse_function, setup_io_handler=True, logger_name=None):
    """
    Initialize common application startup tasks.

    This function handles the repetitive startup tasks common to most simtools applications:

    - Parse command line arguments and configuration
    - Set up logging with appropriate level
    - Optionally initialize IOHandler

    Parameters
    ----------
    parse_function : Callable
        Function that parses configuration and returns (args_dict, db_config) tuple.
        This should be the application's _parse() function.
    setup_io_handler : bool, optional
        Whether to initialize and return an IOHandler instance. Default is True.
    logger_name : str, optional
        Name for the logger. If None, uses the root logger. Default is None.

    Returns
    -------
    args_dict : dict
        Parsed command line arguments and configuration.
    db_config : dict
        Database configuration dictionary.
    logger : logging.Logger
        Configured logger instance.
    io_handler_instance : io_handler.IOHandler or None
        IOHandler instance if setup_io_handler=True, None otherwise.

    Examples
    --------
    Basic usage in an application:

    .. code-block:: python

        def main():
            app_context = startup_application(_parse)

            # Application-specific code follows
            app_context.logger.info("Starting application")
            # ... rest of application logic

    Usage without IOHandler:

    .. code-block:: python

        def main():
            app_context = startup_application(_parse, setup_io_handler=False)

            # Application-specific code follows
            app_context.logger.info("Starting application")
            # ... rest of application logic
    """
    args_dict, db_config = parse_function()

    if logger_name:
        logger = logging.getLogger(logger_name)
    else:
        logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    io_handler_instance = io_handler.IOHandler() if setup_io_handler else None

    return ApplicationContext(
        args=args_dict,
        db_config=db_config,
        logger=logger,
        io_handler=io_handler_instance,
    )


def get_application_label(file_path):
    """
    Get application label from file path.

    This is a convenience function to extract the application name from __file__.

    Parameters
    ----------
    file_path : str
        The __file__ variable from the calling application.

    Returns
    -------
    str
        Application label (filename without extension).

    Examples
    --------
    .. code-block:: python

        def main():
            label = get_application_label(__file__)
            # label will be the filename without .py extension
    """
    return Path(file_path).stem



#!/usr/bin/python3

"""
Run several simtools applications using a configuration file.

Allows to run several simtools applications with a single configuration file, which includes
both the name of the simtools application and the configuration for the application.

This application is used for model parameter setting workflows.
Strong assumptions are applied on the directory structure for input and output files of
applications, for details see the CTAO Simulation Model Parameter Setting
`repository <https://gitlab.cta-observatory.org/cta-science/simulations/simulation-model/\
simulation-model-parameter-setting>`_.

.. tip::
   Browse the CTAO Simulation Model Parameter Setting
   `repository <https://gitlab.cta-observatory.org/cta-science/simulations/simulation-model/\
simulation-model-parameter-setting>`_
   for a list of examples and templates for configuration files.

For simplified configuration, a placeholder called ``__SETTING_WORKFLOW__`` can be used in the
configuration file. This placeholder will be replaced with the directory below ``input``
(example: configuration file is in ``input/LSTN-design/num_gains/20250214T134800/config.yml``,
then the placeholder will be replaced with ``LSTN-design/num_gains/20250214T134800``).
This will also be the directory for any output generated by the application.

Run time environments can be defined in the configuration file using the ``runtime_environment``
block. The following example shows how to define a runtime environment:

.. code-block:: yaml

     runtime_environment:
     image: ghcr.io/gammasim/simtools-prod-sim-telarray-240927-corsika-77550-bernlohr-1.68-\
prod6-baseline-qgs2-no_opt:20250716-122341
     network: simtools-mongo-network
     environment_file: ./.env
     container_engine: podman
     options:
     - '--arch amd64'

If the ``ignore_runtime_environment`` flag is set, the application will run in the current
environment, ignoring any definitions in the configuration file.

The database configuration and setting of other environment variables is done as
described in :ref:`environment-variables`.

Example
-------

Run the application with the configuration file ``config_file_name``:

.. code-block:: console

     simtools-run-application --configuration_file config_file_name

Run the application with the configuration file ``config_file_name``, but skipping all steps except
step 2 and 3 (useful for debugging):

.. code-block:: console

     simtools-run-application --configuration_file config_file_name --steps 2 3

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.runners import simtools_runner


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Run simtools applications using a configuration file.",
        usage="simtools-run-application --config_file config_file_name",
    )

    config.parser.add_argument(
        "--configuration_file",
        help="Application configuration.",
        type=str,
        required=True,
        default=None,
    )
    config.parser.add_argument(
        "--steps",
        type=int,
        nargs="+",
        help="List of steps to be execution (e.g., '--steps 7 8 9'; do not specify to run all).",
    )
    config.parser.add_argument(
        "--ignore_runtime_environment",
        action="store_true",
        help="Ignore the runtime environment and run the application in the current environment.",
        default=False,
    )
    return config.initialize(db_config=True)


def main():
    """Run several simtools applications using a configuration file."""
    app_context = startup_application(_parse, setup_io_handler=False)

    simtools_runner.run_applications(app_context.args, app_context.db_config, app_context.logger)


if __name__ == "__main__":
    main()
